#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
File Name：     data_dectection.py
Author :        weiguang
date：          2021/6/18
Description :
"""
from wgcpy.utils.ext_fn import *
import os

logger = init_logger()

class DectectDF:
    def __init__(self, df):
        if not isinstance(df, pd.DataFrame):
            Exception("the type of df must be DataFrame!")
        self.df = df

    @staticmethod
    def _get_top_values(series, top=5, reverse=False):
        '''
        获取Top对应的值
        :param series: pd.Series
        :param top: int
        :param reverse: bool
        :return: pd.Series
        '''
        itype = 'top'
        counts = series.value_counts()
        counts = list(zip(counts.index, counts, counts.divide(series.size)))

        if reverse:
            counts.reverse()
            itype = 'bottom'

        template = "{0[0]}:{0[2]:.2%}"
        indexs = [itype + str(i + 1) for i in range(top)]
        values = [template.format(counts[i]) if i < len(counts) else None for i in range(top)]
        return pd.Series(values, index=indexs)

    @staticmethod
    def _get_describe(series, percentiles=None):
        '''
        获取数据表描述性统计结果
        :param series: pd.Series
        :param percentiles: list or ndarray
        :return: pd.DataFrame
        '''
        if percentiles is None:
            percentiles = [0.25, 0.5, 0.75]
        d = series.describe(percentiles)
        return d.drop('count')

    @staticmethod
    def _cal_na(series, special_value_dict=None):
        '''
        计算列的缺失占比
        :param series: pd.Series
        :param special_value_dict:
        :return: tuple
        '''
        if not special_value_dict:
            series = series.replace(special_value_dict)
        n = series.isnull().sum()
        return n, "{0:.2%}".format(n / series.size)

    @staticmethod
    def _is_numeric(series):
        '''
        数值型判断
        :param series: pd.Series
        :return: bool
        '''
        return series.dtype.kind in 'ifc'

    def detect(self, special_value_dict=None, output=None):
        '''
        数据表探查
        :param special_value_dict: dict
        :param output, str, the dir of save result
        :return: pd.DataFrame
        '''
        if special_value_dict is None:
            special_value_dict = {}
        if not isinstance(special_value_dict, dict):
            raise Exception("the type of the special_value_dict must be dict!")

        logger.info("start detect dataframe >>>>>>>>>")
        rows = []
        for name, series in self.df.items():
            numeric_index = ['mean', 'std', 'min', '1%', '10%', '50%', '75%', '90%', '99%', 'max']
            discrete_index = ['top1', 'top2', 'top3', 'top4', 'top5', 'bottom5', 'bottom4', 'bottom3', 'bottom2',
                              'bottom1']
            details_index = [numeric_index[i] + '_or_' + discrete_index[i] for i in range(len(numeric_index))]

            if self._is_numeric(series=series):
                des = self._get_describe(
                    series=series,
                    percentiles=[.01, .1, .5, .75, .9, .99]
                )
                des = des.tolist()
            else:
                top5 = self._get_top_values(series=series, top=5)
                bottom5 = self._get_top_values(series=series, top=5, reverse=True)
                des = top5.tolist() + bottom5[::-1].tolist()

            na_cnt, na_pct = self._cal_na(series=series, special_value_dict=special_value_dict)
            row = pd.Series(
                index=['type', 'size', 'missing', 'unique'] + details_index,
                data=[series.dtype, series.size, na_pct, series.nunique()] + des
            )
            row.name = name
            rows.append(row)
        des_df = pd.DataFrame(rows)
        if output:
            if not os.path.isdir(output):
                raise ValueError("output must be dir!")
            des_df.to_csv(os.path.join(output, 'des_df.csv'))
        return des_df

# @timecount()
# def describe_df(df, special_value_dict=None, plot=False):
#     '''
#     数据统计性描述
#     :param df: DataFrame
#     :param special_value_dict: dict
#     :param plot: bool
#     :return:
#     '''
#     df_copy = df.copy()
#     if not isinstance(df, pd.DataFrame):
#         Exception("df must be dataframe!")
#     if not isinstance(special_value_dict, dict):
#         Exception("special_value_dict must be dict!")
#     if special_value_dict:
#         df_copy = df_copy.replace(special_value_dict)
#
#     df_des = pd.concat([
#         (df_copy.shape[0] - df_copy.count()).to_frame('na_count'),
#         (1 - df_copy.count() / df_copy.shape[0]).to_frame('na_rate'),
#         df_copy.dtypes.to_frame('col_type'),
#         pd.DataFrame({
#             'unique_num': [len(df_copy[col].unique()) for col in df_copy]
#         }, index=df_copy.columns)
#     ], axis=1, sort=False)
#     df_des_detail = df_copy.describe().round(2).T
#
#     if plot:
#         plt.plot(range(len(df_des.index)),df_des['na_rate'].sort_values())
#         plt.title("Na Rate")
#         plt.show()
#
#     del df_copy
#     gc.collect()
#     return df_des,df_des_detail



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
@Project:       wgcpy
@File Name:     config.py
@Author:        weiguang
@Date:          2021/6/24
"""

CONFIG = {
    "na_threshold": 0.95,
    "correlation_threshold": 0.95,
    "importance_cumsum_threshold": 0.95,
    "params": {
        "n_estimators": 500,
        "max_depth": 3,
        "learning_rate": 0.1,
        "boosting_type": "gbdt",
        "importance_type": "gain",
        "n_jobs": -1
    },
    "kfold": "StratifiedKFold",
    "groups": None,
    "categorical_feature": None,
    "n_splits": 5,
    "incre_params":{
            "max_depth": 3,
            "learning_rate": 0.1,
            "num_boost_round": 200,
            "metrics": "auc",
            "verbose_eval": 200,
            "verbose": -1,
            "early_stopping_rounds": 100,
            "seed": 10
    },
    "total_iter": 100,
    "step": 2,
    "auc_interval": 0.001
}


#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
@Project:       wgcpy
@File Name:     main.py
@Author:        weiguang
@Date:          2021/6/29
"""
from sklearn.model_selection import train_test_split
from wgcpy.config import CONFIG
from wgcpy.preprocessing.eda import *
from wgcpy.preprocessing.data_dectection import DectectDF
from wgcpy.featureSelector.selector import *
from wgcpy.featureSelector.cal_iv_psi import *
from wgcpy.model.gen_pmml_model import *
from wgcpy.utils.ext_fn import *

pd.options.display.max_columns = 20
pd.options.display.max_rows = 20

def run(credit_data, numeric_feats, category_feats):
    # 数据EDA
    plot_feature_boxplot(credit_data, numeric_feats, save=True)
    plot_feature_distribution(credit_data, numeric_feats,
                              label="flag", sub_col=3, save=True)
    plot_category_countplot(credit_data, category_feats, label="flag",
                            sub_col=5, save=True, figsize=(20,12))
    plot_corr(credit_data, numeric_feats+['flag'], mask=True, save=True)

    # 数据分布
    with timer('detect dataframe'):
        dec = DectectDF(credit_data)
        df_des = dec.detect(special_value_dict={-999:np.nan},
                            output=r"./result")

    # 计算IV
    with timer("cal iv"):
        iv_details = cal_total_var_iv(credit_data,
                                      numeric_feats=numeric_feats,
                                      category_feats=category_feats,
                                      target='flag',
                                      max_interval=10,
                                      method='tree')
        fig = plot_bin_woe(binx=iv_details[iv_details['variable'] == 'credit.amount'],
                           title=None,
                           display_iv=True,
                           save=True)
        iv_details.to_csv(r'./result/iv_details.csv', index=False)

    # 计算psi
    with timer('cal psi'):
        except_array = credit_data['credit.amount'][:500]
        accept_arry = credit_data['credit.amount'][500:]
        psi_df = numeric_var_cal_psi(except_array,
                                     accept_arry,
                                     bins=10,
                                     bucket_type='bins',
                                     detail=True)
        psi_df.to_csv(r'./result/psi.csv')

    # 特征筛选&细筛
    with timer("cal cv score"):
        groups = credit_data['status.of.existing.checking.account']
        config = {
            "na_threshold": 0.95,
            "correlation_threshold": 0.6,
            "importance_cumsum_threshold": 0.95,
            "params": {
                "n_estimators": 200,
                "max_depth": 2,
                "learning_rate": 0.1,
                "boosting_type": "gbdt",
                "importance_type": "gain",
                "n_jobs": -1
            },
            "kfold": "StratifiedKFold",
            "groups": None,
            "categorical_feature": category_feats,
            "n_splits": 5,
            "incre_params": None,
            "total_iter": 20,
            "step": 1,
            "auc_interval": None
        }
        fs = FeatureSelector(data=credit_data,
                             target='flag',
                             base_features=numeric_feats+category_feats)
        fs.identify_all(config=config)
        fs.plot_feature_importance()
        fs.result_save(output=r"./result/feats_seletor_result.xlsx")

    # PMML建模与评估
    with timer("PMML model build"):
        trn_x, tes_x, y_trn, y_tes = train_test_split(credit_data,
                                                      credit_data['flag'],
                                                      test_size=0.2)
        pmml_model = genPMMLModel(data=trn_x,
                                  target="flag",
                                  base_features=numeric_feats+category_feats)

        pmml_model.make_pipeline_model(numeric_feature=numeric_feats,
                                       category_feature=category_feats,
                                       model_type='lgb',
                                       param_dict=config['params'])

        predict = pmml_model.evaluate(data=tes_x,
                                      target="flag")
        pmml_model.persist(base_dir="result",
                           model_name="credit")


if __name__ == "__main__":
    # 数据读取
    credit_data = pd.read_csv(r'data/germancredit.csv')
    numeric_feats = credit_data.select_dtypes(include=['int64','float64']).columns.tolist()
    category_feats = list(set(credit_data.columns) - set(numeric_feats + ['creditability']))
    credit_data["flag"] = credit_data["creditability"].replace({"good": 0, "bad": 1})

    # 运行
    run(credit_data=credit_data,
        category_feats=category_feats,
        numeric_feats=numeric_feats)



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
File Name：     chi_merge.py
Author :        weiguang
date：          2021/6/21
Description :
"""
import warnings
from wgcpy.utils.ext_fn import *

def split_data(df, col, num_split, special_attributes=None):
    """
    细粒度转粗粒度
    :param df: 按照col排序后的数据集
    :param col: 待分箱的变量
    :param num_split: 切分的组别数
    :param special_attributes: 在切分数据集的时候，某些特殊值需要排除在外
    :return: splitPoint: 初始化数据分箱的节点
    """
    if special_attributes is None:
        special_attributes = []
    df2 = df.copy()
    if len(special_attributes) > 0:
        df2 = df.loc[~df.col.isin(special_attributes)]

    N = df2.shape[0]
    n = int(N / num_split)
    splitPointIndex = [i * n for i in range(1, num_split)]
    rawValues = sorted(list(df2[col]))
    splitPoint = [rawValues[i] for i in splitPointIndex]
    splitPoint = sorted(list(set(splitPoint)))

    aa = pd.Series(splitPoint)
    if (aa[0] == 0.0) & (aa.shape[0] == 1):
        num_split = 1000
        n = int(N / num_split)
        splitPointIndex = [i * n for i in range(1, num_split)]
        rawValues = sorted(list(df2[col]))
        splitPoint = [rawValues[i] for i in splitPointIndex]
        splitPoint = sorted(list(set(splitPoint)))
    return splitPoint


def assign_group(x, bins):
    '''
    区间映射
    :param x: float or int
    :param bins: list or tuple, the cut bins
    :return: float or int, the value of
    '''
    N = len(bins)
    if x <= min(bins):
        return min(bins)
    elif x > max(bins):
        return max(bins)
    else:
        for i in range(N - 1):
            if bins[i] < x <= bins[i + 1]:
                return bins[i + 1]


def bin_bad_rate(df, col, target, grantRateIndicator=0):
    """
    用于计算col中数值对应的坏样本的占比情况
    :param df: 需要计算好坏比率的数据集
    :param col: 需要计算好坏比率的特征
    :param target: 好坏标签
    :param grantRateIndicator: 1返回总体的坏样本率，0不返回
    :return: 每箱的坏样本率，以及总体的坏样本率（当grantRateIndicator＝＝1时）
    """
    regroup = df.groupby(col, as_index=False).agg({target: ['count', 'sum']})
    regroup.columns = [col, 'total', 'bad']
    regroup = regroup.assign(bad_rate=lambda x: x.bad / x.total)
    dicts = dict(zip(regroup[col], regroup['bad_rate']))
    if grantRateIndicator == 0:
        return dicts, regroup
    N = sum(regroup['total'])
    B = sum(regroup['bad'])
    overallRate = B * 1.0 / N
    return dicts, regroup, overallRate


def cal_chi2(df, total_col, bad_col):
    '''
    计算卡方值
    :param df: pd.DataFrame, 包含全部样本总计与坏样本总计的数据框
    :param total_col: str, 全部样本的个数
    :param bad_col: str, 坏样本的个数
    :return: float, the chi2 value
    '''
    df2 = df.copy()
    badRate = sum(df2[bad_col]) * 1.0 / sum(df2[total_col])
    if badRate in [0, 1]:
        return 0
    df2 = df2.assign(good=lambda x: x.total - x.bad)
    goodRate = sum(df2['good']) * 1.0 / sum(df2[total_col])
    df2 = df2.assign(
        badExpected=lambda x: x.total * badRate,
        goodExpected=lambda x: x.total * goodRate
    )
    badCombined = zip(df2['badExpected'], df2[bad_col])
    goodCombined = zip(df2['goodExpected'], df2['good'])
    badChi = [(i[0] - i[1]) ** 2 / i[0] for i in badCombined]
    goodChi = [(i[0] - i[1]) ** 2 / i[0] for i in goodCombined]
    chi2 = sum(badChi) + sum(goodChi)
    return chi2


def assign_bin(x, cutOffPoints, special_attributes):
    '''
    bin分箱映射
    :param x: float
    :param cutOffPoints: list
    :param special_attributes: list
    :return: str
    '''
    cutOffPoints2 = [i for i in cutOffPoints if i not in special_attributes]
    numBin = len(cutOffPoints2)
    if x in special_attributes:
        i = special_attributes.index(x) + 1
        return 'Bin {}'.format(0 - i)
    if x <= cutOffPoints2[0]:
        return 'Bin 0'
    elif x > cutOffPoints2[-1]:
        return 'Bin {}'.format(numBin)
    else:
        for ii in range(0, numBin):
            if cutOffPoints2[ii] < x <= cutOffPoints2[ii + 1]:
                return 'Bin {}'.format(ii + 1)


def bad_rate_merge(df, col, cutOffPoints, target, special_attributes=None):
    '''
    将badRate0/1合并分箱节点
    :param df: pd.DataFrame, 包含全部样本总计与坏样本总计的数据框
    :param col: str, 计算好坏样本占比的列
    :param cutOffPoints: list, 分箱节点
    :param target: str, 目标变量
    :param special_attributes: list, 剔除的特殊值
    :return: list, 分箱节点
    '''
    if special_attributes is None:
        special_attributes = []
    while True:
        df2 = df.copy()
        df2['temp_bin'] = df2[col].apply(lambda x: assign_bin(x, cutOffPoints, special_attributes))
        binBadRate, regroup = bin_bad_rate(df2, 'temp_bin', target=target)
        indexForBad01 = regroup[regroup['bad_rate'].isin([0, 1])]['temp_bin'].tolist()
        if len(indexForBad01) == 0:
            return cutOffPoints

        bin = indexForBad01[0]
        # 移除最后一箱
        if bin == max(regroup['temp_bin']):
            cutOffPoints = cutOffPoints[:-1]
            if len(cutOffPoints) == 0:
                return np.nan
        # 移除第一箱
        elif bin == min(regroup['temp_bin']):
            cutOffPoints = cutOffPoints[1:]
            if len(cutOffPoints) == 0:
                return np.nan
        else:
            # 和前一箱进行合并，并且计算卡方值
            currentIndex = list(regroup['temp_bin']).index(bin)
            prevIndex = list(regroup['temp_bin'])[currentIndex - 1]
            df3 = df2.loc[df2['temp_bin'].isin([prevIndex, bin])]
            binBadRate, df2b = bin_bad_rate(df3, 'temp_bin', target)
            chisq1 = cal_chi2(df2b, 'total', 'bad')
            # 和后一箱进行合并，并且计算卡方值
            laterIndex = list(regroup['temp_bin'])[currentIndex + 1]
            df3b = df2.loc[df2['temp_bin'].isin([laterIndex, bin])]
            binBadRate, df2b = bin_bad_rate(df3b, 'temp_bin', target)
            chisq2 = cal_chi2(df2b, 'total', 'bad')
            if chisq1 < chisq2:
                cutOffPoints.remove(cutOffPoints[currentIndex - 1])
            else:
                cutOffPoints.remove(cutOffPoints[currentIndex])

        # 完成合并之后，需要再次计算新的分箱准则下，每箱是否同时包含好坏样本
        df2['temp_bin'] = df2[col].apply(lambda x: assign_bin(x, cutOffPoints, special_attributes))
        binBadRate, regroup = bin_bad_rate(df2, 'temp_bin', target)
        [minBadRate, maxBadRate] = [min(binBadRate.values()),
                                    max(binBadRate.values())]
        if minBadRate > 0 and maxBadRate < 1:
            break
    return cutOffPoints


def cal_chi_merge(df, col, target, max_interval, min_bin_pct=0.01, special_attributes=None):
    '''
    计算卡方分箱节点
    :param df: pd.DataFrame
    :param col: str, the variable of dataframe
    :param target: str, the target of dataframe
    :param max_interval: int, the max interval of variable
    :param min_bin_pct: float, the min percent of bin
    :param special_attributes: list or tuple, remove values of Series
    :return: list, the cutPoints of variable
    '''
    if special_attributes is None:
        special_attributes = []
    check_type(data=df, special_attributes=special_attributes)

    col_level = sorted(list(set(df[col])))
    N_distinct = len(col_level)
    if N_distinct <= max_interval:
        print(f"the number of {col} unique values less then {max_interval}!")
        warnings.warn("warnings!", UserWarning)
        return col_level[:-1]
    else:
        if len(special_attributes) >= 1:
            df2 = df[~df[col].isin(special_attributes)]
        else:
            df2 = df.copy()
        N_distinct = len(list(set(df2[col])))

        if N_distinct > 100:
            split_x = split_data(df2, col, num_split=100)
            df2['temp'] = df2[col].map(lambda x: assign_group(x=x, bins=split_x))
            if len(df2['temp'].unique()) == 1:
                return np.nan
        else:
            df2['temp'] = df2[col]

        binBadRate, regroup, overallRate = bin_bad_rate(df=df2, col='temp', target=target, grantRateIndicator=1)
        col_level = sorted(list(set(df2['temp'])))
        groupIntervals = [[i] for i in col_level]

        # 步骤：建立循环，不断合并最优的相邻两个组别，直到：
        # 1，最终分裂出来的分箱数<＝预设的最大分箱数
        # 2，每箱的占比不低于预设值（可选）
        # 3，每箱同时包含好坏样本
        # 如果有特殊属性，那么最终分裂出来的分箱数＝预设的最大分箱数－特殊属性的个数
        split_intervals = max_interval - len(special_attributes)
        if split_intervals == 1:
            return np.nan
        while len(groupIntervals) > split_intervals:
            chisqList = []
            for k in range(len(groupIntervals) - 1):
                temp_group = groupIntervals[k] + groupIntervals[k + 1]
                df2b = regroup.loc[regroup['temp'].isin(temp_group)]
                chisq = cal_chi2(df2b, 'total', 'bad')
                chisqList.append(chisq)
            best_comnbined = chisqList.index(min(chisqList))
            groupIntervals[best_comnbined] = groupIntervals[best_comnbined] + groupIntervals[best_comnbined + 1]
            groupIntervals.remove(groupIntervals[best_comnbined + 1])
        groupIntervals = [sorted(i) for i in groupIntervals]
        cutOffPoints = [max(i) for i in groupIntervals[:-1]]

        df2['temp_bin'] = df2['temp'].apply(lambda x: assign_bin(x=x, cutOffPoints=cutOffPoints,
                                                                 special_attributes=special_attributes))
        binBadRate, regroup = bin_bad_rate(df2, 'temp_bin', target)
        [minBadRate, maxBadRate] = [min(binBadRate.values()),
                                    max(binBadRate.values())]
        if minBadRate == 0 or maxBadRate == 1:
            cutOffPoints = bad_rate_merge(df, col, cutOffPoints, target)

        # 检查分箱后的最小占比
        if min_bin_pct > 0:
            groupValues = df2['temp'].apply(lambda x: assign_bin(x, cutOffPoints, special_attributes))
            df2['temp_bin'] = groupValues
            valueCounts = groupValues.value_counts().to_frame()
            N = sum(valueCounts['temp'])
            valueCounts['pcnt'] = valueCounts['temp'].apply(
                lambda x: x * 1.0 / N)
            valueCounts = valueCounts.sort_index()
            minPcnt = min(valueCounts['pcnt'])

            # 当最小箱占比不满足预设占比或者分箱节点数小于等于2时,合箱停止
            while minPcnt < min_bin_pct and len(cutOffPoints) > 2:
                indexForMinPcnt = valueCounts[valueCounts['pcnt'] == minPcnt].index.tolist()[0]
                # 占比最小箱是最后一箱
                if indexForMinPcnt == max(valueCounts.index):
                    cutOffPoints = cutOffPoints[:-1]
                # 占比最小箱是第一箱
                elif indexForMinPcnt == min(valueCounts.index):
                    cutOffPoints = cutOffPoints[1:]
                # 占比最小箱是中间箱
                else:
                    # 和前一箱合并计算卡方值
                    currentIndex = list(valueCounts.index).index(indexForMinPcnt)
                    prevIndex = list(valueCounts.index)[currentIndex - 1]
                    df3 = df2.loc[df2['temp_bin'].isin([prevIndex, indexForMinPcnt])]
                    binBadRate, df2b = bin_bad_rate(df3, 'temp_bin', target)
                    chisq1 = cal_chi2(df2b, 'total', 'bad')
                    # 和后一箱合并计算卡方值
                    laterIndex = list(valueCounts.index)[currentIndex + 1]
                    df3b = df2.loc[df2['temp_bin'].isin([laterIndex, indexForMinPcnt])]
                    binBadRate, df2b = bin_bad_rate(df3b, 'temp_bin', target)
                    chisq2 = cal_chi2(df2b, 'total', 'bad')
                    # 比较卡方值, 将较小的卡方值进行合箱
                    if chisq1 < chisq2:
                        cutOffPoints.remove(cutOffPoints[currentIndex - 1])
                    else:
                        cutOffPoints.remove(cutOffPoints[currentIndex])
                df2['temp_bin'] = df2['temp'].apply(lambda x: assign_bin(x, cutOffPoints, special_attributes))
                valueCounts = df2['temp_bin'].value_counts().to_frame()
                valueCounts['pcnt'] = valueCounts['temp'].apply(lambda x: x * 1.0 / N)
                valueCounts = valueCounts.sort_index()
                minPcnt = min(valueCounts['pcnt'])

            if cutOffPoints is np.nan:
                return np.nan
            return cutOffPoints


'''
-----------------------------
单调性合并相关函数
------------------------------
'''

def feature_monotone(x):
    '''
    判断不满足单调性的值对应index
    :param x: list, the cut of bins
    :return: dict
    '''
    monotone = [(x[i] < x[i + 1]) and (x[i] < x[i - 1]) or (x[i] > x[i + 1]) and (x[i] > x[i - 1]) for i in
                range(1, len(x) - 1)]
    index_of_monotone = [i + 1 for i in range(len(monotone)) if monotone[i]]
    return {'count_of_nonmonotone': monotone.count(True),
            'index_of_nonmonotone': index_of_monotone}


def bad_rate_monotone(df, sortByVar, target, special_attributes=None):
    '''
    判断某变量的坏账率是否是单调
    :param df: pd.DataFrame
    :param sortByVar: str
    :param target: str, the target of dataframe
    :param special_attributes: list or tuple
    :return: bool
    '''
    if special_attributes is None:
        special_attributes = []

    df2 = df.loc[~df[sortByVar].isin(special_attributes)]
    if len(set(df2[sortByVar])) <= 2:
        return True
    regroup = bin_bad_rate(df2, sortByVar, target)[1]
    combined = zip(regroup['total'], regroup['bad'])
    badRate = [x[1] * 1.0 / x[0] for x in combined]
    badRateNotMonotone = feature_monotone(badRate)['count_of_nonmonotone']
    if badRateNotMonotone > 0:
        return False
    else:
        return True


def monotone_merge(df, target, col):
    def merge_maxtrix(m, i, j, k):
        """
        :param i: 合并第i行
        :param m: 需要合并行的矩阵
        :param j: 合并第i行
        :param k: 删除第k行
        :return: 合并后的矩阵
        """
        m[i, :] = m[i, :] + m[j, :]
        m = np.delete(m, k, axis=0)
        return m

    def merge_adjacent_rows(i, bad_by_bin_current, bins_list_current, not_monotone_count_current):
        """
        :param i: 需要将第i行与前、后的行分别进行合并，比较哪种合并方案最佳。判断准则是，合并后非单调性程度减轻，且更加均匀
        :param bad_by_bin_current:合并前的分箱矩阵，包括每一箱的样本个数、坏样本个数和坏样本率
        :param bins_list_current: 合并前的分箱方案
        :param not_monotone_count_current:合并前的非单调性元素个数
        :return:分箱后的分箱矩阵、分箱方案、非单调性元素个数和衡量均匀性的指标balance
        """
        i_prev = i - 1
        i_next = i + 1
        bins_list = bins_list_current.copy()
        bad_by_bin = bad_by_bin_current.copy()
        not_monotone_count = not_monotone_count_current

        # 合并方案a：将第i箱与前一箱进行合并
        bad_by_bin2a = merge_maxtrix(bad_by_bin.copy(), i_prev, i, i)
        bad_by_bin2a[i_prev, -1] = bad_by_bin2a[i_prev, -2] / bad_by_bin2a[i_prev, -3]
        not_monotone_count2a = feature_monotone(bad_by_bin2a[:, -1])['count_of_nonmonotone']

        # 合并方案b：将第i行与后一行进行合并
        bad_by_bin2b = merge_maxtrix(bad_by_bin.copy(), i, i_next, i_next)
        bad_by_bin2b[i, -1] = bad_by_bin2b[i, -2] / bad_by_bin2b[i, -3]
        not_monotone_count2b = feature_monotone(bad_by_bin2b[:, -1])['count_of_nonmonotone']
        balance = ((bad_by_bin[:, 1] / N).T * (bad_by_bin[:, 1] / N))[0, 0]
        balance_a = ((bad_by_bin2a[:, 1] / N).T * (bad_by_bin2a[:, 1] / N))[0, 0]
        balance_b = ((bad_by_bin2b[:, 1] / N).T * (bad_by_bin2b[:, 1] / N))[0, 0]

        # 满足下述2种情况时返回方案a：（1）方案a能减轻非单调性而方案b不能；
        # （2）方案a和b都能减轻非单调性，但是方案a的样本均匀性优于方案b
        if (not_monotone_count2a < not_monotone_count_current) and \
                (not_monotone_count2b >= not_monotone_count_current) or \
                (not_monotone_count2a < not_monotone_count_current) and \
                (not_monotone_count2b < not_monotone_count_current) and \
                (balance_a < balance_b):
            bins_list[i_prev] = bins_list[i_prev] + bins_list[i]
            bins_list.remove(bins_list[i])
            bad_by_bin = bad_by_bin2a
            not_monotone_count = not_monotone_count2a
            balance = balance_a
        # 同样地，满足下述2种情况时返回方案b：
        # （1）方案b能减轻非单调性而方案a不能；
        # （2）方案a和b都能减轻非单调性，但是方案b的样本均匀性优于方案a
        elif (not_monotone_count2a >= not_monotone_count_current) and \
                (not_monotone_count2b < not_monotone_count_current) or \
                (not_monotone_count2a < not_monotone_count_current) and \
                (not_monotone_count2b < not_monotone_count_current) and \
                (balance_a > balance_b):
            bins_list[i] = bins_list[i] + bins_list[i_next]
            bins_list.remove(bins_list[i_next])
            bad_by_bin = bad_by_bin2b
            not_monotone_count = not_monotone_count2b
            balance = balance_b
        # 如果方案a和b都不能减轻非单调性，返回均匀性更优的合并方案
        else:
            if balance_a < balance_b:
                bins_list[i] = bins_list[i] + bins_list[i_next]
                bins_list.remove(bins_list[i_next])
                bad_by_bin = bad_by_bin2b
                not_monotone_count = not_monotone_count2b
                balance = balance_b
            else:
                bins_list[i] = bins_list[i] + bins_list[i_next]
                bins_list.remove(bins_list[i_next])
                bad_by_bin = bad_by_bin2b
                not_monotone_count = not_monotone_count2b
                balance = balance_b
        return {'bins_list': bins_list, 'bad_by_bin': bad_by_bin,
                'not_monotone_count': not_monotone_count,
                'balance': balance}

    N = df.shape[0]
    badrate_bin, bad_by_bin = bin_bad_rate(df=df, col=col, target=target)
    bins = list(bad_by_bin[col])
    bins_list = [[i] for i in bins]
    badRate = sorted(badrate_bin.items(), key=lambda x: x[0])
    badRate = [i[1] for i in badRate]
    not_monotone_count, not_monotone_position = feature_monotone(badRate)['count_of_nonmonotone'], \
                                                feature_monotone(badRate)['index_of_nonmonotone']
    # 迭代地寻找最优合并方案，终止条件是:当前的坏样本率已经单调，或者当前只有2箱
    while not_monotone_count > 0 and len(bins_list) > 2:
        all_possible_merging = []
        for i in not_monotone_position:
            merge_adjacent_rows = merge_adjacent_rows(
                i, np.mat(bad_by_bin), bins_list, not_monotone_count)
            all_possible_merging.append(merge_adjacent_rows)
        balance_list = [i['balance'] for i in all_possible_merging]
        not_monotone_count_new = [i['not_monotone_count'] for i in all_possible_merging]
        # 如果所有的合并方案都不能减轻当前的非单调性，就选择更加均匀的合并方案
        if min(not_monotone_count_new) >= not_monotone_count:
            best_merging_position = balance_list.index(min(balance_list))
        # 如果有多个合并方案都能减轻当前的非单调性，也选择更加均匀的合并方案
        else:
            better_merging_index = [
                i for i in range(len(not_monotone_count_new))
                if not_monotone_count_new[i] < not_monotone_count
            ]
            better_balance = [balance_list[i] for i in better_merging_index]
            best_balance_index = better_balance.index(min(better_balance))
            best_merging_position = better_merging_index[best_balance_index]
        bins_list = all_possible_merging[best_merging_position]['bins_list']
        bad_by_bin = all_possible_merging[best_merging_position]['bad_by_bin']
        not_monotone_count = all_possible_merging[best_merging_position]['not_monotone_count']
        not_monotone_position = FeatureMonotone(bad_by_bin[:, 3])['index_of_nonmonotone']
    return bins_list


def cutpoint_brm(data, var, label, cp, special_attributes=None):
    '''
    检验每一箱中坏样本分布是否单调，不单调需要将分箱进行上下合并
    :param data: pd.DataFrame
    :param var: str, the variable of dataframe
    :param label: str, the label of dataframe
    :param cp: list, the cut bins of variable
    :param special_attributes: list r tuple
    :return: list
    '''
    data = data.copy()
    var_cutoff = dict()
    col1 = str(var) + '_Bin'
    cp = sorted(list(cp))
    if special_attributes is not None:
        special_attributes = []
    # 将col1按照cp进行分组映射
    data[col1] = data[var].map(
        lambda x: assign_bin(
            x, cp, special_attributes=special_attributes
        )
    )
    binBadRate, regroup = bin_bad_rate(data, col1, label)
    [minBadRate, maxBadRate] = [min(binBadRate.values()),
                                max(binBadRate.values())]
    if minBadRate == 0 or maxBadRate == 1:
        if len(binBadRate) == 2:
            return np.nan
        else:
            # 坏账率分箱后检验与合箱
            cp = bad_rate_merge(data, var, cp, label)
            return cp
    data[col1] = data[var].map(
        lambda x: assign_bin(x, cp, special_attributes=special_attributes)
    )
    var_cutoff[var] = cp
    brm = bad_rate_monotone(data, col1, label, special_attributes=special_attributes)
    # 分箱后坏账率不单调，则继续进行合箱
    if not brm:
        bin_merge = monotone_merge(data, label, col1)
        removed_index = []
        for bin in bin_merge:
            if len(bin) > 1:
                indices = [int(b.replace('Bin ', '')) for b in bin]
                removed_index = removed_index + indices[0:-1]
        removed_point = [cp[k] for k in removed_index]
        for p in removed_point:
            cp.remove(p)
        var_cutoff[var] = cp
        data[col1] = data[var].map(
            lambda x: assign_bin(
                x, cp, special_attributes=special_attributes
            )
        )
    return cp



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
@Project:       wgcpy
@File Name:     dz_eval.py
@Author:        weiguang
@Date:          2021/6/28
"""
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from sklearn.metrics import roc_curve,auc,precision_recall_curve,\
                            average_precision_score,roc_auc_score
from sklearn.model_selection import learning_curve,validation_curve
from sklearn.metrics import confusion_matrix
from sklearn.utils.multiclass import unique_labels
import seaborn as sns

def plot_confusion_matrix(y_true, y_pred, classes,
                          normalize=False,
                          title=None,
                          cmap=plt.cm.Blues):
    '''
    混淆矩阵绘制
    :param y_true: ndarray, the true label
    :param y_pred: ndarray, the predict label
    :param classes: list, the class of label
    :param normalize: bool
    :param title: str
    :param cmap:
    :return:
    '''
    if not title:
        if normalize:
            title = 'Normalized confusion matrix'
        else:
            title = 'Confusion matrix, without normalization'

    # Compute confusion matrix
    cm = confusion_matrix(y_true, y_pred)
    # Only use the labels that appear in the data
    classes = classes[unique_labels(y_true, y_pred)]
    if normalize:
        cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]
        print("Normalized confusion matrix")
    else:
        print('Confusion matrix, without normalization')
    print(cm)

    fig, ax = plt.subplots()
    im = ax.imshow(cm, interpolation='nearest', cmap=cmap)
    ax.figure.colorbar(im, ax=ax)
    # We want to show all ticks...
    ax.set(xticks=np.arange(cm.shape[1]),
           yticks=np.arange(cm.shape[0]),
           xticklabels=classes, yticklabels=classes,
           title=title,
           ylabel='True label',
           xlabel='Predicted label')

    # Rotate the tick labels and set their alignment.
    plt.setp(ax.get_xticklabels(), rotation=45, ha="right",
             rotation_mode="anchor")

    # Loop over data dimensions and create text annotations.
    fmt = '.2f' if normalize else 'd'
    thresh = cm.max() / 2.
    for i in range(cm.shape[0]):
        for j in range(cm.shape[1]):
            ax.text(j, i, format(cm[i, j], fmt),
                    ha="center", va="center",
                    color="white" if cm[i, j] > thresh else "black")
    fig.tight_layout()
    plt.show()


def plot_ks_curve(preds, labels, is_score=False, n=100,
                  return_value=True, return_graph=False, return_table=False):
    '''
    KS绘制
    :param preds: ndarray, the predict labels
    :param labels: ndarray, the true labels
    :param is_score: bool, whether the predict labels is socre
    :param n: int, the bins of split
    :param return_value: bool, whether need to return value
    :param return_graph: bool, whether need to return graph
    :param return_table: bool, whether need to return detail table
    :return:
    '''
    ksds = pd.DataFrame({'bad': labels, 'pred': preds})
    ksds['good'] = 1 - ksds.bad

    if is_score:
        ksds1 = ksds.sort_values(by=['pred', 'bad'], ascending=[True, True])
    else:
        ksds1 = ksds.sort_values(by=['pred', 'bad'], ascending=[False, True])
    ksds1.index = range(len(ksds1.pred))
    ksds1['cumsum_good1'] = 1.0 * ksds1.good.cumsum() / sum(ksds1.good)
    ksds1['cumsum_bad1'] = 1.0 * ksds1.bad.cumsum() / sum(ksds1.bad)

    if is_score:
        ksds2 = ksds.sort_values(by=['pred', 'bad'], ascending=[True, False])
    else:
        ksds2 = ksds.sort_values(by=['pred', 'bad'], ascending=[False, False])
    ksds2.index = range(len(ksds2.pred))
    ksds2['cumsum_good2'] = 1.0 * ksds2.good.cumsum() / sum(ksds2.good)
    ksds2['cumsum_bad2'] = 1.0 * ksds2.bad.cumsum() / sum(ksds2.bad)

    # ksds1 ksds2 -> average
    ksds = ksds1[['cumsum_good1', 'cumsum_bad1']]
    ksds['cumsum_good2'] = ksds2['cumsum_good2']
    ksds['cumsum_bad2'] = ksds2['cumsum_bad2']
    ksds['cumsum_good'] = (ksds['cumsum_good1'] + ksds['cumsum_good2']) / 2
    ksds['cumsum_bad'] = (ksds['cumsum_bad1'] + ksds['cumsum_bad2']) / 2

    # ks
    ksds['ks'] = ksds['cumsum_bad'] - ksds['cumsum_good']
    ksds['tile0'] = range(1, len(ksds.ks) + 1)
    ksds['tile'] = 1.0 * ksds['tile0'] / len(ksds['tile0'])

    qe = list(np.arange(0, 1, 1.0 / n))
    qe.append(1)
    qe = qe[1:]

    ks_index = pd.Series(ksds.index)
    ks_index = ks_index.quantile(q=qe)
    ks_index = np.ceil(ks_index).astype(int)
    ks_index = list(ks_index)

    ksds = ksds.loc[ks_index]
    ksds = ksds[['tile', 'cumsum_good', 'cumsum_bad', 'ks']]
    ksds0 = np.array([[0, 0, 0, 0]])
    ksds = np.concatenate([ksds0, ksds], axis=0)
    ksds = pd.DataFrame(ksds,
                        columns=['tile', 'cumsum_good', 'cumsum_bad', 'ks'])

    ks_value = ksds.ks.max()
    ks_pop = ksds.tile[ksds.ks.idxmax()]
    if return_graph:
        plt.figure(figsize=(6,4))
        print('ks_value is ' + str(np.round(ks_value, 4)) + ' at pop = ' + str(
        np.round(ks_pop, 4)))
        # chart
        plt.plot(ksds.tile, ksds.cumsum_good, label='cum_good',
                 color='blue', linestyle='-', linewidth=2)
        plt.plot(ksds.tile, ksds.cumsum_bad, label='cum_bad',
                 color='red', linestyle='-', linewidth=2)
        plt.plot(ksds.tile, ksds.ks, label='ks',
                 color='green', linestyle='-', linewidth=2)
        plt.axvline(ks_pop, color='gray', linestyle='--')
        plt.axhline(ks_value, color='green', linestyle='--')
        plt.axhline(ksds.loc[ksds.ks.idxmax(), 'cumsum_good'], color='blue',
                    linestyle='--')
        plt.axhline(ksds.loc[ksds.ks.idxmax(), 'cumsum_bad'], color='red',
                    linestyle='--')
        plt.title('KS=%s ' % np.round(ks_value, 4) +
                  'at Pop=%s' % np.round(ks_pop, 4), fontsize=15)
        plt.show()
    if return_value:
        return ks_value
    elif return_table:
        return ksds, ks_value
    else:
        assert (return_table or return_graph or return_value),'请设置参数为True'
        return None


def plot_roc_curve(y_true, y_predict, return_value=True, return_graph=False):
    '''
    ROC曲线
    :param y_true: ndarray, the true labels
    :param y_predict: ndarray, the predict labels
    :param return_value: bool, whether need to return value
    :param return_graph: bool, whether need to return graph
    :return:
    '''
    fpr, tpr, threshold = roc_curve(y_true, y_predict)
    roc_auc = auc(fpr, tpr)
    if return_graph:
        plt.figure()
        plt.figure(figsize=(6, 4))
        lw=2
        plt.plot(fpr, tpr, color='darkorange', lw=lw, label='ROC curve (area = %0.2f)' % roc_auc)
        plt.plot([0, 1], [0, 1], color='navy', lw=lw, linestyle='--')
        plt.xlim([0.0, 1.0])
        plt.ylim([0.0, 1.05])
        plt.xlabel('False Positive Rate')
        plt.ylabel('True Positive Rate')
        plt.title('Roc Curve')
        plt.legend(loc="lower right")
        plt.show()
    if return_value:
        return roc_auc


# 绘制验证曲线
def plot_validation_curve(estimator, title, X, y, param_name, param_range, ylim=(0, 1), cv=5, scoring='auc'):
    '''
    用于绘制验证曲线
    :param estimator: model
    :param title: str
    :param X: pd.DataFrame
    :param y: pd.Series or ndarray
    :param param_name: str
    :param param_range: list
    :param ylim: tuple (0, 1)
    :param cv: int
    :param scoring: str, the type of evaluation
    :return: pd.DataFrame
    '''
    train_scores, test_scores = validation_curve(
        estimator, X, y, param_name=param_name, param_range=param_range,
        cv=cv, scoring=scoring)
    train_scores_mean = np.mean(train_scores, axis=1)
    train_scores_std = np.std(train_scores, axis=1)
    test_scores_mean = np.mean(test_scores, axis=1)
    test_scores_std = np.std(test_scores, axis=1)

    plt.title(title)
    plt.xlabel(param_name)
    plt.ylabel(scoring)
    if ylim is not None:
        plt.ylim(*ylim)
    lw = 2
    plt.semilogx(param_range, train_scores_mean, label="Training score",
                 color="darkorange", lw=lw)
    plt.fill_between(param_range, train_scores_mean - train_scores_std,
                     train_scores_mean + train_scores_std, alpha=0.2,
                     color="darkorange", lw=lw)
    plt.semilogx(param_range, test_scores_mean, label="Cross-validation score",
                 color="navy", lw=lw)
    plt.fill_between(param_range, test_scores_mean - test_scores_std,
                     test_scores_mean + test_scores_std, alpha=0.2,
                     color="navy", lw=lw)
    plt.legend(loc="best")
    plt.show()
    return pd.DataFrame({'train_scores_mean': train_scores_mean, 'train_scores_std': train_scores_std,
                         'test_scores_mean': test_scores_mean, 'test_scores_std': test_scores_std})


def plot_learning_curve(estimator, title, X, y, ylim=(0,1), cv=None,
                        train_sizes=np.linspace(.1, 1.0, 5), scoring='roc_auc'):
    '''
    绘制学习曲线
    :param estimator: model
    :param title: str
    :param X: pd.DataFrame
    :param y: pd.Series or ndarray
    :param ylim: tuple or list, default is (0,1)
    :param cv: int
    :param train_sizes: list
    :param scoring: the type of evaluation
    :return: pd.DataFrame
    '''
    plt.figure()
    plt.title(title)
    if ylim is not None:
        plt.ylim(*ylim)
    plt.xlabel("Training examples")
    plt.ylabel(scoring)
    train_sizes, train_scores, test_scores = learning_curve(
        estimator, X, y, cv=cv, n_jobs=6, scoring=scoring, train_sizes=train_sizes)
    train_scores_mean = np.mean(train_scores, axis=1)
    train_scores_std = np.std(train_scores, axis=1)
    test_scores_mean = np.mean(test_scores, axis=1)
    test_scores_std = np.std(test_scores, axis=1)
    plt.grid()

    plt.fill_between(train_sizes, train_scores_mean - train_scores_std,
                     train_scores_mean + train_scores_std, alpha=0.1,
                     color="r")
    plt.fill_between(train_sizes, test_scores_mean - test_scores_std,
                     test_scores_mean + test_scores_std, alpha=0.1, color="g")
    plt.plot(train_sizes, train_scores_mean, 'o-', color="r",
             label="Training score")
    plt.plot(train_sizes, test_scores_mean, 'o-', color="g",
             label="Cross-validation score")

    plt.legend(loc="best")
    plt.show()
    return pd.DataFrame({'train_scores_mean': train_scores_mean, 'train_scores_std': train_scores_std,
                         'test_scores_mean': test_scores_mean, 'test_scores_std': test_scores_std})



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
@Project:       wgcpy
@File Name:     eda.py
@Author:        weiguang
@Date:          2021/6/18
"""
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns

def plot_feature_boxplot(df, feats, figsize=(40,12), sub_col=2, save=False):
    '''
    绘制箱型图
    :param df: DataFrame
    :param feats: list or ndarray
    :param figsize: tuple
    :param sub_col: int
    :param save: bool
    :return:
    '''
    if not isinstance(df, pd.DataFrame):
        raise Exception("The type of df must be DataFrame")
    plt.subplots(figsize=figsize)

    for i,feat in enumerate(feats):
        ax1 = plt.subplot(sub_col, np.ceil(len(feats)/sub_col), i+1)
        plt.xlim(df[feat].min(), df[feat].max()*1.1)
        sns.boxplot(x=df[feat], ax=ax1)
    if save:
        plt.savefig("./pic/plot_feature_boxplot.png", format="png")
    plt.show()


def plot_feature_distribution(df, feats, label='flag', sub_col=5, figsize=(35,20), save=False):
    '''
    绘制特征分布图
    :param df: DataFrame
    :param feats: list or ndarray
    :param label: string, the label of df
    :param sub_col: int
    :param figsize: tuple
    :param save: bool
    :return:
    '''
    if not isinstance(df, pd.DataFrame):
        raise Exception("The type of df must be DataFrame")
    label_v = np.unique(df[label])
    d1 = df[df[label].isin([label_v[1]])]
    d0 = df[df[label].isin([label_v[0]])]
    plt.figure(figsize=figsize)

    for i,feat in enumerate(feats):
        ax1 = plt.subplot(sub_col, np.ceil(len(feats)/sub_col), i+1)
        sns.kdeplot(d1[feat], label=f'{label_v[1]}', bw=1.5, ax=ax1)
        sns.kdeplot(d0[feat], label=f'{label_v[0]}', bw=1.5, ax=ax1)
        plt.xlabel(feat, fontsize=12)
        locs, labels = plt.xticks()
        plt.tick_params(axis='x', which='major', labelsize=10, pad=-6)
        plt.tick_params(axis='y', which='major', labelsize=10)
    if save:
        plt.savefig("./pic/plot_feature_distribution.png", format="png")
    plt.show()


def plot_category_countplot(df, feats, label="flag", sub_col=2, save=False, figsize=(12,8)):
    '''
    绘制类别型特征分布
    :param df: DataFrame
    :param feats: list or ndarray
    :param label: string, the label of df
    :param sub_col: int
    :param save: bool
    :param figsize: tuple
    :return:
    '''
    if not isinstance(df, pd.DataFrame):
        raise Exception("The type of df must be DataFrame")
    sns.set(style='darkgrid', color_codes=True)
    plt.figure()
    plt.subplots(figsize=figsize)

    for i,feat in enumerate(feats):
        plt.subplot(sub_col, np.ceil(len(feats)/sub_col), i+1)
        sns.countplot(x=feat, hue=label, data=df, palette='Set3')

    if save:
        plt.savefig("./pic/plot_category_countplot.png", format="png")
    plt.show()


def plot_corr(df, feats, figsize=(12,12), mask=False, save=False):
    '''
    绘制相关性图
    :param df: pd.DataFrame
    :param feats: list or array
    :param figsize: tuple
    :param mask: bool
    :param save: bool
    :return:
    '''
    if not isinstance(df, pd.DataFrame):
        raise Exception("The type of df must be DataFrame")
    corr_df = df.loc[:,feats].corr()
    colormap = plt.cm.viridis
    plt.figure(figsize=figsize)
    plt.title('Heatmap', y=1.05, size=15)
    if mask:
        mask = np.triu(np.ones_like(corr_df,dtype=np.bool))
    sns.heatmap(corr_df, linewidths=0.1, vmax=1.0, mask=mask,
                square=True, cmap=colormap, linecolor='white', annot=True)
    if save:
        plt.savefig("./pic/plot_corr.png", format="png")
    plt.show()

def plot_bin_woe(binx, title=None, display_iv=False, figsize=(8, 6), save=False):
    '''
    绘制woe
    :param binx: pd.DataFrame, the details of bins
    :param title: str, the title of the figure
    :param display_iv: bool, whether to show iv
    :param figsize: tuple, the size of figsize
    :param save: bool
    :return: figure
    '''
    y_right_max = np.ceil(binx['pct_1_row'].max() * 10)
    if y_right_max % 2 == 1:
        y_right_max = y_right_max + 1
    if y_right_max - binx['pct_1_row'].max() * 10 <= 0.3:
        y_right_max = y_right_max+2
    y_right_max = y_right_max/10
    if y_right_max > 1 or y_right_max <= 0 or y_right_max is np.nan or y_right_max is None:
        y_right_max = 1

    y_left_max = np.ceil(binx['pct_bin'].max() * 10) / 10
    if y_left_max > 1 or y_left_max <= 0 or y_left_max is np.nan or y_left_max is None:
        y_left_max = 1

    title_string = binx.loc[0, 'variable'] + "  (iv:" + str(binx.loc[0, 'IV']) + ")" \
                    if display_iv else binx.loc[0, 'variable']
    title_string = title + '-' + title_string if title is not None else title_string
    binx['good_distr'] = binx['num_0'] / np.sum(binx['num_01'])
    binx['bad_distr'] = binx['num_1'] / np.sum(binx['num_01'])
    ind = np.arange(len(binx.index))
    width = 0.35
    fig, ax1 = plt.subplots(figsize=figsize)
    ax2 = ax1.twinx()
    # ax1
    p1 = ax1.bar(ind, binx['good_distr'], width, color=(24 / 254, 192 / 254, 196 / 254))
    p2 = ax1.bar(ind, binx['bad_distr'], width, bottom=binx['good_distr'], color=(246 / 254, 115 / 254, 109 / 254))
    for i in ind:
        ax1.text(i, binx.loc[i, 'pct_bin'] * 1.02,
                 str(round(binx.loc[i, 'pct_bin'] * 100, 1)) + '%, ' + str(binx.loc[i, 'num_01']), ha='center')
    # ax2
    ax2.plot(ind, binx['pct_1_row'], marker='o', color='blue')
    for i in ind:
        ax2.text(i, binx.loc[i, 'pct_1_row'] * 1.02, str(round(binx.loc[i, 'pct_1_row'] * 100, 1)) + '%', color='blue',
                 ha='center')
    # settings
    ax1.set_ylabel('Bin count distribution', fontdict={"fontsize":10})
    ax2.set_ylabel('Bad probability', fontdict={"fontsize":10, "color":"blue"})
    ax1.set_yticks(np.arange(0, y_left_max + 0.2, 0.2))
    ax2.set_yticks(np.arange(0, y_right_max + 0.2, 0.2))
    ax2.tick_params(axis='y', colors='blue')
    plt.xticks(ind, binx.index)
    for xtick in ax1.get_xticklabels():
        xtick.set_rotation(30)

    plt.title(title_string, loc='left')
    plt.legend((p2[0], p1[0]), ('bad', 'good'), loc='upper right')
    if save:
        plt.savefig(f"./pic/{binx.loc[0, 'variable']}_woe.png", format="png")
    plt.show()
    return fig



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
@Project:       wgcpy
@File Name:     gen_pmml_model.py
@Author:        weiguang
@Date:          2021/6/28
"""
import joblib
from wgcpy.utils.ext_fn import *
from .dz_eval import *
from lightgbm import LGBMClassifier
from xgboost import XGBClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.ensemble import VotingClassifier
from sklearn2pmml import sklearn2pmml
from sklearn.impute import SimpleImputer
from sklearn_pandas import DataFrameMapper,CategoricalImputer
from sklearn2pmml.decoration import ContinuousDomain,CategoricalDomain
from sklearn2pmml.pipeline import PMMLPipeline
from sklearn_pandas import gen_features
from sklearn.pipeline import FeatureUnion
from sklearn.preprocessing import StandardScaler,LabelBinarizer

logger = init_logger()

class genPMMLModel:
    def __init__(self, data, target, base_features):
        self.base_features = base_features
        self.train_data = data.loc[:, self.base_features]
        self.target_data = data.loc[:, target]
        self.param_dict = None
        self.pipeline_model = None
        self.numeric_feature = None
        self.category_feature = None

    def _gen_model(self, model_type, param_dict):
        '''
        生成模型
        :param model_type: str, model type
        :param param_dict: dict, the param dict of model
        :return: model
        '''
        if isinstance(param_dict, dict):
            self.param_dict = param_dict

        if model_type == 'lgb':
            model = LGBMClassifier().set_params(**self.param_dict.get('lightgbm', {}))
        elif model_type == 'voting':
            lgb_model = LGBMClassifier().set_params(**self.param_dict.get('lightgbm', {}))
            xgb_model = XGBClassifier().set_params(**self.param_dict.get('xgboost', {}))
            lr_model = LogisticRegression().set_params(**self.param_dict.get('lr', {}))
            model = VotingClassifier(estimators=[('lightgbm', lgb_model), ('xgboost', xgb_model), ('lr', lr_model)],
                                     voting='soft', weights=np.ones(3), n_jobs=3)
        else:
            raise ValueError('the type must be lgb or voting! ')
        return model


    def make_pipeline_model(self, numeric_feature, category_feature, model_type, param_dict):
        '''
        PMML模型构建
        :param numeric_feature: list or ndarray, the numeric feature of train data
        :param category_feature: list or ndarray, the category feature of train data
        :param model_type: str, the model type of train, only support lgb and voting!
        :param param_dict: dict, the param of model
        :return:
        '''
        if isinstance(numeric_feature, (list, np.ndarray)):
            self.numeric_feature = numeric_feature

        if isinstance(category_feature, (list, np.ndarray)):
            self.category_feature = category_feature

        if isinstance(param_dict, dict):
            self.param_dict = param_dict

        model = self._gen_model(model_type=model_type, param_dict=self.param_dict)
        feature_def = gen_features(
            columns=self.category_feature,
            classes=[CategoricalDomain, CategoricalImputer, LabelBinarizer]
            )
        mapper_numerical = DataFrameMapper([
                (self.numeric_feature,[ContinuousDomain(with_data=False),
                                       SimpleImputer(missing_values=np.nan, fill_value=-999, strategy="constant"),
                                       StandardScaler()])
        ])
        mapper_category = DataFrameMapper(feature_def)

        if len(self.category_feature) == 0:
            mapper = mapper_numerical
        else:
            mapper = FeatureUnion([('mapper_numerical',mapper_numerical), ('mapper_category',mapper_category)])
        self.pipeline_model = PMMLPipeline([
            ('mapper', mapper),
            ('classifier', model)
        ])
        logger.info(f'numeric features cnt: {len(self.numeric_feature)}, \t'
                    f'category features cnt: {len(self.category_feature)}')
        logger.info(f'train data shape: {self.train_data.shape}, \t'          
                    f'target data shape: {self.target_data.shape}')
        self.pipeline_model.fit(self.train_data, self.target_data)
        self.pipeline_model.verify(self.train_data.sample(n=10))


    def evaluate(self, data, target):
        '''
        模型评估
        :param data: pd.DataFrame, the data need to evaluated
        :param target: str, the name of target label
        :return: ndarray, the predict result of data
        '''
        ts_data = data.loc[:, self.base_features]
        ts_target = data.loc[:, target]
        predict_ts = self.pipeline_model.predict_proba(ts_data)[:, 1]
        auc = plot_roc_curve(y_true=ts_target, y_predict=predict_ts, return_graph=True)
        ks = plot_ks_curve(preds=predict_ts, labels=ts_target, return_graph=True)
        logger.info(f"auc:{auc}, ks:{ks}")
        return predict_ts


    def persist(self, base_dir, model_name):
        '''
        模型保存
        :param base_dir: str, the path of save model
        :param model_name: str, model name
        :return:
        '''
        joblib.dump(self.pipeline_model, os.path.join(base_dir, f"{model_name}.pkl"), compress=3)
        sklearn2pmml(self.pipeline_model, os.path.join(base_dir, f"{model_name}.pmml"), with_repr=True)




#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
File Name：     cut_bins.py
Author :        weiguang
date：          2021/6/21
Description :
"""
from sklearn.tree import DecisionTreeClassifier
from .chi_merge import *
import warnings
warnings.filterwarnings("ignore")

TREE_UNDEFINED = -2

def interpolate_binning(data, var, special_attributes=None):
    '''
    插值分箱
    :param data: pd.DataFrame
    :param var: str, the var of dataframe
    :param special_attributes: list or tuple, remove values of Series
    :return: list, the bins of variable
    '''
    if special_attributes is None:
        special_attributes = []
    check_type(data=data, special_attributes=special_attributes)

    bin_s = data[var].loc[~data[var].isin(special_attributes)].sort_values()
    check_single_value(series=bin_s)
    if bin_s.isna().sum() > 0:
        raise ValueError(f"detect nan values in {var}")
    value_list = list(bin_s.value_counts().sort_index().index)
    cp = [(value_list[i] + value_list[i + 1]) / 2 for i in np.arange(len(value_list) - 1)]

    if not check_unique(cp):
        cp = sorted(list(set(cp)))
    return cp


def quantile_binning(data, var, max_interval=10, special_attributes=None):
    '''
    等频分箱
    :param data: pd.DataFrame
    :param var: str, the var of DataFrame
    :param max_interval: int, the max bins of Series
    :param special_attributes: list or tuple, remove values of Series
    :return: list, the bins of variable
    '''
    if special_attributes is None:
        special_attributes = []

    check_type(data=data, special_attributes=special_attributes)
    bin_s = data[var].loc[~data[var].isin(special_attributes)].sort_values()
    check_single_value(series=bin_s)
    if bin_s.isna().sum() > 0:
        raise ValueError(f"detect nan values in {var}")

    unique_c = len(bin_s.value_counts())
    if unique_c < max_interval:
        print(f"the unique of {var} less then {max_interval}")
        warnings.warn(f"warnings!",UserWarning)
        max_interval = unique_c
    cp = [bin_s.quantile(i) for i in np.linspace(0, 1, max_interval + 1)[1:-1]]

    if not check_unique(cp):
        cp = sorted(list(set(cp)))
    return cp


def distance_binning(data, var, max_interval=10, special_attributes=None):
    '''
    等距分箱
    :param data: pd.DataFrame
    :param var: str, the var of DataFrame
    :param max_interval: int, the max bins of Series
    :param special_attributes: list or tuple, remove values of Series
    :return: list, the bins of variable
    '''
    if special_attributes is None:
        special_attributes = []

    check_type(data=data, special_attributes=special_attributes)
    bin_s = data[var].loc[~data[var].isin(special_attributes)].sort_values()
    check_single_value(series=bin_s)
    if bin_s.isna().sum() > 0:
        raise ValueError(f"detect nan values in {var}")

    unique_c = len(bin_s.value_counts())
    if unique_c < max_interval:
        print(f"the unique of {var} less then {max_interval}")
        warnings.warn("warnings!", UserWarning)
        max_interval = unique_c
    cp = list(np.linspace(bin_s.min(), bin_s.max(), max_interval + 1, endpoint=True)[1:-1])

    if not check_unique(cp):
        cp = sorted(list(set(cp)))
    return cp


def mix_binning(data, var, max_interval=10, special_attributes=None):
    '''
    混合分箱
    :param data: pd.DataFrame
    :param var: str, the var of DataFrame
    :param max_interval: int, the max bins of Series
    :param special_attributes: list or tuple
    :return: list, the bins of variable
    '''
    if special_attributes is None:
        special_attributes = []

    check_type(data=data, special_attributes=special_attributes)
    bin_s = data[var].loc[~data[var].isin(special_attributes)].sort_values()
    if np.sum(bin_s.isna()) > 0:
        raise ValueError("detect nan values in {0}".format(var))

    unique_c = len(bin_s.value_counts())
    if unique_c < max_interval:
        print(f"the unique of {var} less then {max_interval}")
        warnings.warn("warnings!", UserWarning)
        max_interval = unique_c
    quantile_cp = [bin_s.quantile(i) for i in np.linspace(0,1,max_interval+1)][1:-1]
    distance_cp = list(np.linspace(quantile_cp[0], quantile_cp[-1], max_interval-1, endpoint=True)[1:-1])
    cp = [quantile_cp[0]] + distance_cp + [quantile_cp[-1]]

    if not check_unique(cp):
        cp = sorted(list(set(cp)))
    return cp


def tree_binning(data, var, target, special_attributes=None, tree_params=None):
    '''
    决策树分箱
    :param data: pd.DataFrame
    :param var: str, the var of dataframe
    :param target: the target of dataframe
    :param special_attributes: list or tuple, remove values of Series
    :param tree_params: dict, the params of DecisionTree
    :return: list, the bins of variable
    '''
    if tree_params is None:
        tree_params = {'max_leaf_nodes': 10, 'criterion': 'gini', 'min_samples_leaf': 0.01}

    if special_attributes is None:
        special_attributes = []
    check_type(data=data, special_attributes=special_attributes)
    bin_data = data[[var, target]].loc[~data[var].isin(special_attributes)]
    check_single_value(bin_data[var])
    if (np.sum(bin_data[var].isna()) > 0) or (np.sum(bin_data[target].isna()) > 0):
        raise ValueError(f"detect nan values in {var}")

    Dtree = DecisionTreeClassifier().set_params(**tree_params)
    Dtree.fit(bin_data[var].values.reshape(-1,1), bin_data[target].values.reshape(-1, 1))
    cp = sorted(Dtree.tree_.threshold[Dtree.tree_.threshold != TREE_UNDEFINED])
    if len(cp) == 0:
        raise ValueError(f"detect empty cp for {var} in tree bins!")
    return cp


def chi_binning(data, var, target, max_interval=10, special_attributes=None):
    '''
    卡方分箱
    :param data: pd.DataFrame
    :param var: str, the variable of dataframe
    :param target: str, the target of dataframe
    :param max_interval: int, the max interval of bins
    :param special_attributes: list or tuple
    :return: list, the cut points of variable
    '''
    if special_attributes is None:
        special_attributes = []
    check_type(data=data, special_attributes=special_attributes)
    binning_data = data[[var, target]].loc[~data[var].isin(special_attributes)]

    if (np.sum(binning_data[var].isna()) > 0) or (np.sum(binning_data[target].isna()) > 0):
        raise ValueError("detect nan values in {0}".format([var,target]))

    unique_c = len(binning_data[var].value_counts())
    if unique_c < max_interval:
        print(f"value_counts for {var} is {unique_c}, less than max_interval {max_interval}")
        warnings.warn("warnings!", UserWarning)
        max_interval = unique_c

    cp = cal_chi_merge(data, var, target, max_interval=max_interval, special_attributes=special_attributes)
    if (cp is np.nan) or not cp:
        raise ValueError("detect empty cp for {0} in chi_binning".format([var, target]))
    return cp




#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
@Project:       wgcpy
@File Name:     selector.py
@Author:        weiguang
@Date:          2021/6/23
"""
from pandas import ExcelWriter
from wgcpy.utils.ext_fn import *
from wgcpy.config import CONFIG
import matplotlib.pyplot as plt
from wgcpy.model.gen_model import GenCVModel, IncreaseCVSelector

logger = init_logger()

class FeatureSelector:
    def __init__(self, data, target, base_features):
        self.na_threshold = None
        self.labels = data.loc[:, target]
        self.data = data.loc[:, base_features]

        if not isinstance(data, pd.DataFrame):
            raise Exception("data must be DataFrame!")

        self.one_hot_features = None
        self.correlation_threshold = None
        self.importance_threshold = None
        self.useful_feats = None
        self.drop_feats = None

        self.record_na = None
        self.record_single_value = None
        self.record_collinear = None
        self.record_feats_importance = None
        self.record_cv_result = None
        self.record_increase_result = None
        self.record_increase_feats = None
        self.corr_matrix = None
        self.ops = {}


    def identify_na(self, na_threshold=1):
        '''
        缺失率剔除
        :param na_threshold: float, the threshold of missing rate
        :return:
        '''
        self.na_threshold = na_threshold
        na_rate_series = self.data.isna().sum() / self.data.shape[0]
        self.record_na = pd.DataFrame(data={"feature": na_rate_series.index, "na_fraction":na_rate_series.values})
        self.record_na = self.record_na.sort_values("na_fraction", ascending=False)
        self.record_na['mask'] = self.record_na['na_fraction'] >= na_threshold
        self.ops['drop_na_col'] = self.record_na[self.record_na['mask']]
        logger.info('%d features with greater than %0.2f missing values.\n' % (len(self.ops['drop_na_col']), self.na_threshold))


    def identify_single_value(self):
        '''
        剔除单一值
        :return:
        '''
        unique_counts = self.data.nunique()
        self.record_single_value = pd.DataFrame(data={'feature':unique_counts.index, 'nunique':unique_counts.values})
        self.record_single_value = self.record_single_value.sort_values('nunique', ascending=True)
        self.record_single_value['mask'] = self.record_single_value['nunique'] == 1
        self.ops['drop_single_col'] = self.record_single_value[self.record_single_value['mask']]
        logger.info('%d features with a single unique value.\n' % len(self.ops['drop_single_col']))


    def identify_collinear(self, correlation_threshold=1):
        '''
        剔除相关性
        :param correlation_threshold: float, the correlation threshold
        :return:
        '''
        self.correlation_threshold = correlation_threshold
        self.corr_matrix = self.data.corr()
        upper = self.corr_matrix.where(np.triu(np.ones(self.corr_matrix.shape), k=1).astype(np.bool))
        drop_corr_col = [i for i in upper.columns if np.any(upper[i].abs() > correlation_threshold)]

        self.record_collinear = pd.DataFrame(columns=['drop_feature', 'corr_feature', 'corr_value'])
        for col in drop_corr_col:
            corr_feats = list(upper.index[upper[col].abs() > correlation_threshold])
            corr_values = list(upper[col][upper[col].abs() > correlation_threshold])
            drop_feats = [col for _ in range(len(corr_feats))]
            tmp = pd.DataFrame.from_dict({'drop_feature': drop_feats,
                                          'corr_feature': corr_feats,
                                          'corr_value': corr_values})
            self.record_collinear = self.record_collinear.append(tmp, ignore_index=True)
        self.ops['collinear'] = self.record_collinear
        logger.info('%d features with a correlation magnitude greater than %0.2f.\n' % (len(self.ops['collinear']),
                                                                                        self.correlation_threshold))


    def identify_cv_importance(self, params, kfold, categorical_feature=None, groups=None, n_splits=5):
        '''
        通过cv计算importance
        :param params: dict, the params of estimator
        :param kfold: str, the cv of split data
        :param categorical_feature: list, the category of feature
        :param groups: ndarray or Series
        :param n_splits: int
        :return:
        '''
        gen_model = GenCVModel(train_data=self.data, target_data=self.labels)
        self.record_feats_importance, self.record_cv_result = gen_model.cross_validation(params=params, kfold=kfold,
                                                                                         categorical_feature=categorical_feature,
                                                                                         groups=groups, n_splits=n_splits)
        self.ops['drop_zero_importance'] = self.record_feats_importance[self.record_feats_importance['importance'] == 0]


    def identify_increase_cv_feats(self, total_iter, step, categorical_feature, auc_interval, incre_params=None):
        """
        自动筛选特征，方向为验证集auc提升方向
        :param total_iter: int, the max iterator of traing round
        :param step: int, the numbers of add features
        :param categorical_feature: list, the category of features
        :param auc_interval: int, the miniumn interval of auc increase
        :param incre_params: dict, the params of increase model
        :return:
        """
        increase_model = IncreaseCVSelector(train_data=self.data, target_data=self.labels)
        self.record_increase_result, self.record_increase_feats = increase_model.get_lgb_cv_score(self.record_feats_importance,
                                                                                                  total_iter=total_iter,
                                                                                                  step=step,
                                                                                                  incre_params=incre_params,
                                                                                                  categorical_feature=categorical_feature,
                                                                                                  auc_interval=auc_interval)


    def identify_all(self, config=None):
        '''
        筛选特征
        :param config:
        :return:
        '''
        if config is None:
            config = CONFIG
        self.importance_threshold = config.get("importance_cumsum_threshold", CONFIG["importance_cumsum_threshold"])
        self.identify_na(na_threshold=config.get('na_threshold', CONFIG['na_threshold']))
        self.identify_single_value()
        self.identify_collinear(correlation_threshold=config.get("correlation_threshold", CONFIG["correlation_threshold"]))

        self.drop_feats = list(self.ops['drop_na_col'].feature) + list(self.ops["drop_single_col"].feature)\
                    + list(self.ops["collinear"].drop_feature)
        self.useful_feats = self.data.columns[~self.data.columns.isin(self.drop_feats)]
        logger.info(f"detect the total feats, drop feats: {len(self.drop_feats)}")
        logger.info(f"detect the total feats, userful feats: {len(self.useful_feats)}")

        self.data = self.data.loc[:, self.useful_feats]
        self.identify_cv_importance(
            params=config.get("params", CONFIG['params']),
            kfold=config.get("kfold", CONFIG["kfold"]),
            groups=config.get("groups", CONFIG["groups"]),
            categorical_feature=config.get("categorical_feature", CONFIG["categorical_feature"]),
            n_splits=config.get("n_splits", CONFIG["n_splits"])
        )

        self.record_feats_importance = self.record_feats_importance[self.record_feats_importance['cumsum']
                                                                    < self.importance_threshold]

        logger.info(f"After filtering cumsum importance above {'{:.2%}'.format(self.importance_threshold)}, "
                    f"The number of seleted features is: {len(self.record_feats_importance)} !")

        self.identify_increase_cv_feats(
            total_iter=config.get('total_iter', CONFIG['total_iter']),
            step=config.get("step", CONFIG["step"]),
            categorical_feature=config.get("categorical_feature", CONFIG["categorical_feature"]),
            auc_interval=config.get("auc_interval", CONFIG["auc_interval"])
        )


    def plot_feature_importance(self, n=50, figsize=(20, 12)):
        '''
        绘制特征重要度
        :param n: int, the top-n features
        :param figsize: tuple, the figsize of plot
        :return:
        '''
        plt.figure(figsize=figsize)
        ax = plt.subplot()

        ax.barh(list(reversed(list(self.record_increase_feats.index[:n]))),
                self.record_increase_feats['importance_normalized'].head(n),
                align='center', edgecolor='r')
        ax.set_yticks(list(reversed(list(self.record_increase_feats.index[:n]))))
        ax.set_yticklabels(self.record_increase_feats['feature'].head(n))

        plt.xlabel('Normalized Importance')
        plt.title(f"Feature Importances: {':.2%'.format(sum(self.record_increase_feats['importance_normalized']))}")
        plt.show()


    def result_save(self, output=None):
        '''
        结果文件保存
        :param output: string, the path of save result
        :return:
        '''
        if not output.endswith("xlsx"):
            raise AssertionError("output must be a file and endswith xlsx!")

        with ExcelWriter(output, engine="xlsxwriter", mode="w") as writer:
            self.record_cv_result.to_excel(excel_writer=writer,
                                           sheet_name=u"模型CV-Importance评估结果",
                                           index=True)
            self.record_feats_importance.to_excel(excel_writer=writer,
                                                  sheet_name=u"模型CV-{}Importance".format('{:.2%}'.format(self.importance_threshold)),
                                                  index=True)
            self.record_increase_feats.to_excel(excel_writer=writer,
                                                sheet_name="特征IncreaseCV-Filter评估结果",
                                                index=True)
            self.record_increase_result.to_excel(excel_writer=writer,
                                                 sheet_name="模型IncreaseCV-STEP验证结果",
                                                 index=True)




#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
@Project:       wgcpy
@File Name:     ext_fn.py
@Author:        weiguang
@Date:          2021/6/24
"""
import sys, os
import numpy as np
import pandas as pd
import time
from contextlib import contextmanager
import logging

PACKAGENAME = 'wgcpy'
LOGFILE = os.path.join(os.path.dirname(os.getcwd()), "info.log")

LEVEL = {
    'debug': logging.DEBUG,
    'info': logging.INFO,
    'warning': logging.WARNING,
    'error': logging.ERROR,
    'crit': logging.CRITICAL
}

@contextmanager
def timer(name):
    t0 = time.time()
    yield
    logger.info(f'[{name}] done in {time.time() - t0:.02f} s')

def init_logger(level=LEVEL['info']):
    logger = logging.getLogger(PACKAGENAME)
    logger.setLevel(level=level)
    if not logger.handlers:
        message_format = logging.Formatter(fmt='%(asctime)s %(name)s >>> %(message)s',
                                           datefmt='%Y-%m-%d %H-%M-%S')
        ch_va = logging.StreamHandler(sys.stdout)
        ch_va.setLevel(logging.INFO)
        ch_va.setFormatter(fmt=message_format)
        logger.addHandler(ch_va)

        ch_lg = logging.FileHandler(LOGFILE, encoding='utf-8')
        ch_lg.setLevel(logging.INFO)
        ch_lg.setFormatter(fmt=message_format)
        logger.addHandler(ch_lg)
    return logger


logger = init_logger()
logger.info(f"wgcpy log save path: {LOGFILE}! ")

def is_monotonic(x):
    '''
    判断是否单调
    :param x: list or array or series
    :return: bool
    '''
    dx = np.diff(x)
    return np.all(dx < 0).astype(int) or np.all(dx > 0).astype(int)


def check_unique(x):
    '''
    检测是否唯一值
    :param x:  list or array or series
    :return: bool
    '''
    if len(np.unique(x)) != len(x):
        return False
    else:
        return True


def check_non_intersect(x, y):
    '''
    判断x与y是否存在交集
    :param x:  list or array or series
    :param y: list or array or series
    :return: bool
    '''
    if len(set(x) & set(y)) != 0:
        print("存在交集:%s" % (set(x) & set(y)))
        return False
    else:
        return True


def dict_reverse(d):
    '''
    将字典key与value反转
    :param d: dict
    :return: dict
    '''
    reverse_dict = {}
    for key, value in d.items():
        if value not in reverse_dict:
            reverse_dict[value] = [key]
        else:
            reverse_dict[value].append(key)
    return reverse_dict


def check_single_value(series):
    '''
    检测唯一值
    :param series: pd.Series
    :return: bool
    '''
    if len(series.value_counts()) == 1:
        raise ValueError("The value of Series is single!")


def check_type(data, special_attributes):
    '''
    检测数据类型
    :param data: pd.DataFrame
    :param special_attributes: list or tuple
    :return:
    '''
    if not isinstance(data, pd.DataFrame):
        raise Exception("the type of data must be DataFrame!")
    if not isinstance(special_attributes, (list, tuple)):
        raise Exception("the type of special_attribute must be list or tuple!")


def scale_range(input_array, scaled_min, scaled_max):
    '''
    区间缩放
    :param input_array: array
    :param scaled_min: float
    :param scaled_max: float
    :return: array
    '''
    input_array += -np.min(input_array)
    if scaled_max == scaled_min:
        raise Exception('scaled max equal scaled min, please check expected_array！')
    scaled_slope = np.max(input_array) * 1.0 / (scaled_max - scaled_min)
    input_array /= scaled_slope
    input_array += scaled_min
    return input_array


def generate_counts(arr, breakpoints):
    '''
    Generates counts for each bucket by using the bucket values
    :param arr: ndarray of actual values
    :param breakpoints: list of bucket values
    :return cnt_array: counts for elements in each bucket, length of breakpoints array minus one
    :return score_range_array: bins of cut
    '''

    def count_in_range(arr, low, high, start):
        '''
        Counts elements in array between low and high values)
        :param arr: ndarray of actual values
        :param low: float, the boundray of left bin
        :param high: float, the boundray of right bin
        :param start: bool, the mode of bins
        '''
        if start:
            cnt_in_range = len(np.where(np.logical_and(arr >= low, arr <= high))[0])
        else:
            cnt_in_range = len(np.where(np.logical_and(arr > low, arr <= high))[0])
        return cnt_in_range

    cnt_array = np.zeros(len(breakpoints) - 1)
    score_range_array = [''] * (len(breakpoints) - 1)
    for i in range(1, len(breakpoints)):
        cnt_array[i - 1] = count_in_range(arr, breakpoints[i - 1], breakpoints[i], i == 1)
        if 1 == i:
            score_range_array[i - 1] = '[' + str(round(breakpoints[i - 1], 4)) + ',' + str(
                round(breakpoints[i], 4)) + ']'
        else:
            score_range_array[i - 1] = '(' + str(round(breakpoints[i - 1], 4)) + ',' + str(
                round(breakpoints[i], 4)) + ']'
    return cnt_array, score_range_array


def sub_psi(e_perc, a_perc):
    '''
    cal psi value
    :param e_perc: array, the excepted array
    :param a_perc: array, the actual array
    :return: float, the psi value
    '''
    if a_perc == 0:
        a_perc = 0.00001
    if e_perc == 0:
        e_perc = 0.00001
    psi_v = (e_perc - a_perc) * np.log(e_perc * 1.0 / a_perc)
    return psi_v


class CheckDataType(object):
    def __init__(self, train_data, target_data):
        self.target_data = target_data
        self.train_data = train_data

    def _check_target_data_type(self):
        if not any([isinstance(self.target_data, obj_type) for obj_type in [pd.Series, np.ndarray, list]]):
            raise TypeError(
                '"target" must be "numpy.ndarray" or "Pandas.Series" or "list", got {} instead.'.format(type(self.target_data)))
        try:
            assert len(self.target_data.shape) == 1, 'target must be 1-D. It is {}-D instead.'.format(len(self.target_data.shape))
        except AttributeError:
            logger.info('Cannot determine shape of the {}. '
                        'Type must be "numpy.ndarray" or "Pandas.Series" or "list", got {} instead'.format(name,type(self.target_data)))

    @property
    def format_target(self):
        self._check_target_data_type()
        if isinstance(self.target_data, pd.Series):
            return self.target_data.values
        elif isinstance(self.target_data, np.ndarray):
            return self.target_data
        elif isinstance(self.target_data, list):
            return np.array(self.target_data)
        else:
            raise TypeError('"target" must be "numpy.ndarray" or "Pandas.Series" or "list", got {} instead.'.format(type(self.target_data)))

    @property
    def format_train(self):
        if isinstance(self.train_data, (pd.Series, pd.DataFrame)):
            return self.train_data
        else:
            raise TypeError("train data must be DataFrame or Series!")




#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
@Project:       wgcpy
@File Name:     setup.py
@Author:        weiguang
@Date:          2021/6/29
"""
from setuptools import setup, find_packages

with open(r"./requirements.txt") as f:
    requires = [i for i in f if not i.startswith("#")]

setup(
    name="wgcpy",
    version="1.0.0",
    author="weiguang",
    author_email="1677234597@qq.com",
    description="Data analysis and PMML model framework!",
    packages=find_packages(),
    zip_safe=True,
    install_requires=requires,
    url="https://github.com/whyjust/wgcpy"
)



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
@Project:       wgcpy
@File Name:     cal_iv_psi.py
@Author:        weiguang
@Date:          2021/6/22
"""
from tqdm import tqdm
from wgcpy.utils.ext_fn import *
from wgcpy.bins.cut_bins import *

logger = init_logger()

def numeric_var_binning(data, var, target, bins, delta=1e-6):
    '''
    连续型变量计算iv值
    :param data: pd.DataFrame
    :param var: str, the variable of dataframe
    :param target: str, the target of dataframe
    :param bins: list, the bins of variable
    :param delta: float, the tiny offset of the divisor
    :return:
        bins_df: pd.DataFrame, the detail of bins
        IV: float, the value of IV
        bins: list, the cut bins of variable
    '''
    df = data[[var, target]].copy()
    df['bins'] = pd.cut(x=df[var], bins=bins)
    bins_df = pd.crosstab(df['bins'], df[target])
    bins_df.columns = ['num_0', 'num_1']
    bins_df['num_01'] = bins_df.apply(lambda x: sum(x), axis=1)
    bins_df['variable'] = var
    bins_df = bins_df[['variable', 'num_0', 'num_1', 'num_01']].reset_index()
    bins_df = bins_df.assign(
        pct_0_row=lambda x: x.num_0 / x.num_01,
        pct_1_row=lambda x: x.num_1 / x.num_01,
        pct_0_col=lambda x: x.num_0 / sum(x.num_0),
        pct_1_col=lambda x: x.num_1 / sum(x.num_1),
        pct_bin=lambda x: x.num_01 / sum(x.num_01)
    )
    bins_df['odds'] = bins_df['pct_1_row'] / bins_df['pct_0_row']
    bins_df['woe'] = np.log(bins_df['pct_1_col'] / (bins_df['pct_0_col'] + delta))
    bins_df['miv'] = (bins_df['pct_1_col'] - bins_df['pct_0_col']) * bins_df['woe']
    bins_df['woe'] = bins_df['woe'].replace({np.inf: 0, -np.inf: 0})
    bins_df['miv'] = bins_df['miv'].replace({np.inf: 0, -np.inf: 0})
    IV = bins_df['miv'].sum()
    bins_df['IV'] = [np.round(IV, 4)] * bins_df.shape[0]
    bins_df['is_monotonic'] = [is_monotonic(bins_df['woe'].values)] * bins_df.shape[0]
    return bins_df, IV, bins


def numeric_var_cal_iv(data, var, target, max_interval=10, method='DecisionTree', BRM=False,
                       special_attributes=None, tree_params=None, delta=1e-6):
    '''
    分箱并计算IV值
    :param data: pd.DataFrame
    :param var: str, the variable of dataframe
    :param target: str, the variable of dataframe
    :param max_interval: int
    :param method: the type of cut points
    :param BRM: bool, whether you need to merge according to monotony
    :param special_attributes: list or tuple, the special value of dataframe
    :param tree_params: dict, the params of DecisionTree
    :param delta: float, the tiny offset of the divisor
    :return:
        bins_df: pd.DataFrame, the detail of bins
        IV: float, the value of IV
        bins: list, the cut bins of variable
    '''
    if special_attributes is None:
        special_attributes = []

    logger.info(f"start calculate IV: {var}")
    df = data[[var, target]].copy()
    if method == 'chi':
        cp = chi_binning(df, var, target, max_interval, special_attributes)
    elif method == "tree":
        cp = tree_binning(df, var, target, special_attributes, tree_params)
    elif method == "distance":
        cp = distance_binning(df, var, max_interval, special_attributes)
    elif method == "quantile":
        cp = quantile_binning(df, var, max_interval, special_attributes)
    elif method == "mix":
        cp = mix_binning(df, var, max_interval, special_attributes)
    elif method == 'interpolate':
        cp = interpolate_binning(data, var, special_attributes)
    else:
        method_list = ['chi', 'tree', 'distance', 'quantile', 'mix', 'interpolate']
        raise ValueError(f"Can only use these method:{method_list}!")
    if BRM:
        cp = cutpoint_brm(data, var, target, cp, special_attributes)

    if len(special_attributes) > 1 or (len(special_attributes) == 1 and special_attributes[0] > df[var].min()):
        raise ValueError(f"special_attribute for {0} is {1}, not a empty list or a one element list contains the \
                            smallest value".format(var, special_attributes))
    bins = sorted([-np.inf] + special_attributes + cp + [np.inf])
    bins_df, IV, bins = numeric_var_binning(df, var, target, bins, delta)
    IV = np.round(IV, 4)
    logger.info(f"calculate iv complete! bins cnt:{len(bins)}, IV value: {IV}")
    return bins_df, IV, bins


def numeric_var_woe_transform(data, var, bins_df, bins):
    """
    数值型woe转换
    :param data: pd.DataFrame
    :param var: str, the variable of dataframe
    :param bins_df: pd.DataFrame, the detail of bins
    :param bins: list, the cut bins of variable
    :return: array, the woe array
    """
    df = data[[var]].copy()
    transform_dict = dict(zip(bins_df.index.tolist(), bins_df['woe'].values.tolsit()))
    df['bins'] = pd.cut(x=df[var], bins=bins)
    df['woe'] = df['bins'].relplace(transform_dict)
    return df['woe'].values


def category_var_bins_merge(data, var, target, max_interval=10, method='default', special_attributes=None,
                            tree_params=None):
    '''
    将类别型特征分箱
    :param data: pd.DataFrame
    :param var: str, the variable of dataframe
    :param target: str, the target of dataframe
    :param max_interval: int, the max interval of variable
    :param method: str, the type of cut bins
    :param special_attributes: list or tuple
    :param tree_params: dict, the param of DecisionTree, useful when method is tree
    :return: dict, the bins dict of variable
    '''
    if special_attributes is None:
        special_attributes = []

    df = data[[var, target]][~data[var].isin(special_attributes)].copy()
    bins_df = pd.crosstab(df[var], df[target])
    bins_df.columns = ['num_0', 'num_1']
    bins_df['num_01'] = bins_df.apply(lambda x: sum(x), axis=1)
    bins_df = bins_df.assign(
        pct_0_row=lambda x: x.num_0 / x.num_01,
        pct_1_row=lambda x: x.num_1 / x.num_01
    )
    var_order = list(bins_df.sort_values(by="pct_1_row", ascending=False).index)
    stage1_bins_dict = {}
    for i in var_order:
        stage1_bins_dict[i] = var_order.index(i)
    df[var + '_stage1_bin'] = df[var].replace(stage1_bins_dict)
    stage2_bins_dict = {}
    if method == 'default':
        # 对于default方法 只需要把降序的对应关系稍加处理 把新分组的名称改成Bin_x
        for key in stage1_bins_dict.keys():
            stage2_bins_dict[key] = "Bin_" + "%03d" % (stage1_bins_dict.get(key))
    elif method == 'chi':
        # 对于ChiMerge方法 要把降序后的新分组当成变量传入到卡方分箱当中 然后把划分点拿出来对降序后的组号进行再分组
        cp = chi_binning(df, var + '_stage1_bin', target, max_interval)
        stage2_bins = sorted([-np.inf] + cp + [np.inf])
        stage1_bin_df = pd.DataFrame({'stage1_bin': np.arange(len(stage1_bins_dict))})
        stage1_bin_df['stage2_bin'] = pd.cut(x=stage1_bin_df['stage1_bin'], bins=stage2_bins)
        stage1_bin_df['stage2_bin_final'] = stage1_bin_df['stage2_bin']
        index = 0
        for i in stage1_bin_df['stage2_bin'].value_counts().sort_index().index:
            stage1_bin_df['stage2_bin_final'] = stage1_bin_df['stage2_bin_final'].replace({i: 'Bin_' + "%03d" % index})
            index += 1
        for i in stage1_bins_dict.keys():
            stage2_bins_dict[i] = stage1_bin_df['stage2_bin_final'][stage1_bin_df['stage1_bin']
                                                                    == stage1_bins_dict.get(i)].values[0]
    elif method == 'tree':
        # 对于DecisionTree方法，
        cp = tree_binning(df, var + '_stage1_bin', target, tree_params)
        stage2_bins = sorted([-np.inf] + cp + [np.inf])
        stage1_bin_df = pd.DataFrame({'stage1_bin': np.arange(len(stage1_bins_dict))})
        stage1_bin_df['stage2_bin'] = pd.cut(x=stage1_bin_df['stage1_bin'], bins=stage2_bins)
        stage1_bin_df['stage2_bin_final'] = stage1_bin_df['stage2_bin']
        index = 0
        for i in stage1_bin_df['stage2_bin'].value_counts().sort_index().index:
            stage1_bin_df['stage2_bin_final'] = stage1_bin_df['stage2_bin_final'].replace(
                {i: 'Bin_' + "%03d" % index})
            index += 1
        for i in stage1_bins_dict.keys():
            stage2_bins_dict[i] = stage1_bin_df['stage2_bin_final'][stage1_bin_df['stage1_bin']
                                                                    == stage1_bins_dict.get(i)].values[0]
    else:
        method_list = ['default', 'chi', 'tree']
        raise ValueError(f"Can only use these method:{method_list}!")

    for i in special_attributes:
        stage2_bins_dict[i] = "Bin_special"
    return stage2_bins_dict


def category_var_binning(data, var, target, bins, delta=1e-6):
    '''
    类别型特征IV详情
    :param data: pd.DataFrame
    :param var: str, the variable of dataframe
    :param target: str, the target of dataframe
    :param bins: dict, the bins dict of variable
    :param delta: float, the tiny offset of the divisor
    :return:
        bins_df: pd.DataFrame, the detail of bins
        IV: float, the value of IV
        bins: list, the cut bins of variable
    '''
    df = data[[var, target]].copy()
    reverse_bins = dict_reverse(bins)
    for i in reverse_bins.keys():
        reverse_bins[i] = ','.join(list(map(str, reverse_bins.get(i))))
    df['bins'] = df[var].replace(bins)
    bins_df = df[['bins', target]].groupby(['bins'], as_index=False).agg([np.sum, len])
    bins_df.columns = ['num_1', 'num_01']
    bins_df['num_0'] = bins_df['num_01'] - bins_df['num_1']
    bins_df['variable'] = var
    bins_df = bins_df[['variable', 'num_0', 'num_1', 'num_01']]
    bins_df = bins_df.assign(
        pct_0_row=lambda x: x.num_0 / x.num_01,
        pct_1_row=lambda x: x.num_1 / x.num_01,
        pct_0_col=lambda x: x.num_0 / sum(x.num_0),
        pct_1_col=lambda x: x.num_1 / sum(x.num_1),
        pct_bin=lambda x: x.num_01 / sum(x.num_01)
    )
    bins_df['odds'] = bins_df['pct_1_row'] / bins_df['pct_0_row']
    bins_df['woe'] = np.log(bins_df['pct_1_col'] / (bins_df['pct_0_col'] + delta))
    bins_df['miv'] = (bins_df['pct_1_col'] - bins_df['pct_0_col']) * bins_df['woe']
    bins_df['woe'] = bins_df['woe'].replace({np.inf: 0, -np.inf: 0})
    bins_df['miv'] = bins_df['miv'].replace({np.inf: 0, -np.inf: 0})
    IV = bins_df['miv'].sum()
    bins_df['IV'] = [np.round(IV, 4)] * bins_df.shape[0]
    bins_df['is_monotonic'] = [is_monotonic(bins_df['woe'].values)] * bins_df.shape[0]
    return bins_df, IV, bins


def category_var_cal_iv(data, var, target, max_interval=10, method='default', special_attributes=None,
                        tree_params=None, delta=1e-6):
    '''
    类别型特征计算IV
    :param data: pd.DataFrame
    :param var: str, the variable of dataframe
    :param target: str, the target of dataframe
    :param max_interval: int, the max interval of variable
    :param method: str, the type of cut bins
    :param special_attributes: list or tuple
    :param tree_params: dict, the param of DecisionTree
    :param delta: float, the tiny offset of the divisor
    :return:
        bins_df: pd.DataFrame, the detail of bins
        IV: float, the value of IV
        bins: dict, the cut bins of variable
    '''
    if special_attributes is None:
        special_attributes = []

    logger.info(f"start calculate IV: {var}")
    df = data[[var, target]].copy()
    bins = category_var_bins_merge(df, var, target, max_interval, method, special_attributes, tree_params)
    bins_df, IV, bins = category_var_binning(df, var, target, bins, delta)
    IV = np.round(IV, 4)
    logger.info(f"calculate iv complete! bins cnt:{len(bins)}, IV value: {IV}")
    return bins_df, IV, bins


def cal_total_var_iv(data, numeric_feats, category_feats, target,
                     max_interval=10, method='tree', BRM=False,
                     special_dict=None, tree_params=None):
    '''
    全量特征计算IV
    :param data: pd.DataFrame
    :param numeric_feats: list or ndarray, the numeric feats of data
    :param category_feats: list or ndarray, the category feats of data
    :param target: str, the target of data
    :param max_interval: int, the max interval of variable
    :param method: str, the methed of cut bins, default DecisionTree
    :param BRM: bool, whether to merge bins by monotone
    :param special_dict: dict, the special value of cut bins
    :param tree_params: dict, the tree param of DecisionTree, userful only when method='tree'
    :return: pd.DataFram, the details iv of total variable
    '''
    if special_dict is None:
        special_dict = {}

    iv_details = pd.DataFrame()
    if len(numeric_feats) > 0:
        for feat in tqdm(numeric_feats):
            num_bins_df, num_IV, num_bins = numeric_var_cal_iv(data, feat, target, max_interval=max_interval,
                                                               method=method, BRM=BRM, tree_params=tree_params,
                                                               special_attributes=special_dict.get(feat, None))
            num_bins_df['type'] = 'numeric'
            iv_details = pd.concat([iv_details, num_bins_df])
    if len(category_feats) > 0:
        for feat in tqdm(category_feats):
            cate_bins_df, cate_IV, cate_bins = category_var_cal_iv(data, feat, target, max_interval=max_interval,
                                                                   special_attributes=special_dict.get(feat, None),
                                                                   method=method, tree_params=tree_params)
            cate_bins_df['type'] = 'category'
            iv_details = pd.concat([iv_details, cate_bins_df])
    return iv_details


def category_var_woe_transform(data, var, bins_df, bins):
    '''
    类别型woe转换
    :param data: pd.DataFrame
    :param var: str, the variable of dataframe
    :param bins_df: pd.DataFrame, the detail of bins
    :param bins: dict, the bins dict of variable
    :return: array, the woe array
    '''
    df = data[[var]].copy()
    transform_dict = dict(zip(bins_df.index.tolist(), bins_df['woe'].values.tolist()))
    df['bins'] = df[var].replace(bins)
    df['woe'] = df['bins'].replace(transform_dict)
    return df['woe'].values


def numeric_var_cal_psi(expected_array, actual_array, bins=10, bucket_type='bins', detail=False):
    '''
    :param expected_array: numpy array of original values
    :param actual_array: numpy array of new values, same size as expected
    :param bins: number of percentile ranges to bucket the values into
    :param bucket_type: string, bins or quantiles for choose
    :param detail: bool, whether get the detail table
    :return psi_value: psi
    '''
    logger.info("start calculate psi >>>>>")
    expected_array = pd.Series(expected_array).dropna()
    actual_array = pd.Series(actual_array).dropna()
    if isinstance(list(expected_array)[0], str) or isinstance(list(actual_array)[0], str):
        raise Exception("the value of expected_array or actual_array need to be string!")

    if np.min(expected_array) == np.max(expected_array):
        return -1

    breakpoints = np.arange(0, bins + 1) / bins * 100
    if 'bins' == bucket_type:
        breakpoints = np.linspace(np.min(expected_array), np.max(expected_array), len(breakpoints))
    elif 'quantiles' == bucket_type:
        breakpoints = np.stack([np.percentile(expected_array, b) for b in breakpoints])

    expected_cnt = generate_counts(expected_array, breakpoints)[0]
    expected_percents = expected_cnt / len(expected_array)
    actual_cnt = generate_counts(actual_array, breakpoints)[0]
    actual_percents = actual_cnt / len(actual_array)
    delta_percents = actual_percents - expected_percents
    score_range_array = generate_counts(expected_array, breakpoints)[1]
    sub_psi_array = [sub_psi(expected_percents[i], actual_percents[i]) for i in range(0, len(expected_percents))]

    if detail:
        psi_value = pd.DataFrame()
        psi_value['score_range'] = score_range_array
        psi_value['expecteds'] = expected_cnt
        psi_value['expected(%)'] = expected_percents * 100
        psi_value['actucals'] = actual_cnt
        psi_value['actucal(%)'] = actual_percents * 100
        psi_value['ac - ex(%)'] = delta_percents * 100
        psi_value['actucal(%)'] = psi_value['actucal(%)'].apply(lambda x: round(x, 4))
        psi_value['ac - ex(%)'] = psi_value['ac - ex(%)'].apply(lambda x: round(x, 4))
        psi_value['ln(ac/ex)'] = psi_value.apply(lambda row: np.log((row['actucal(%)'] + 0.00001)
                                                                    / (row['expected(%)'] + 0.00001)), axis=1)
        psi_value['psi'] = sub_psi_array
        psi_value = psi_value.append([{'score_range': 'summary',
                                       'expecteds': sum(expected_cnt),
                                       'expected(%)': 100,
                                       'actucals': sum(actual_cnt),
                                       'actucal(%)': 100,
                                       'ac - ex(%)': np.nan,
                                       'ln(ac/ex)': np.nan,
                                       'psi': np.sum(sub_psi_array)}], ignore_index=True)
        logger.info(
            f"calclate psi complete, psi value: {psi_value.loc[psi_value.score_range == 'summary', 'psi'].values[0]}")
    else:
        psi_value = np.sum(sub_psi_array)
        logger.info(f"calculate psi complete, psi value: {psi_value}")
    return psi_value



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
@Project:       wgcpy
@File Name:     gen_model.py
@Author:        weiguang
@Date:          2021/6/24
"""
import gc
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
from wgcpy.utils.ext_fn import *
from scipy.stats import ks_2samp
import lightgbm as lgb
from sklearn.metrics import roc_auc_score
from sklearn.model_selection import KFold, GroupKFold, StratifiedKFold, TimeSeriesSplit, LeaveOneGroupOut

class GenCVModel(CheckDataType):
    '''
    通过CV方式计算特征重要度
    '''
    def __init__(self, train_data, target_data):
        super(GenCVModel, self).__init__(train_data, target_data)
        self.target_data = self.format_target
        self.train_data = self.format_train
        self.n_splits = None
        self.kfold = None

    def _gen_kfold(self, kfold=None, groups=None):
        '''
        数据集切分
        :param kfold: str, the type of cv
        :param groups: ndarray or Series, the groups of split data, it is useful
                       when kfold is GroupKfold or LeaveOneGroupOut
        :return:
        '''
        if kfold == "GroupKfold":
            groups = self._check_groups(groups=groups)
            k_fold = GroupKFold(n_splits=self.n_splits)
            self.kfold = k_fold.split(self.train_data, self.target_data, groups=groups)

        elif kfold == "StratifiedKFold":
            k_fold = StratifiedKFold(n_splits=self.n_splits, shuffle=True, random_state=10)
            self.kfold = k_fold.split(self.train_data, self.target_data)

        elif kfold == "KFold":
            k_fold = KFold(n_splits=self.n_splits, shuffle=True, random_state=10)
            self.kfold = k_fold.split(self.train_data)

        elif kfold == "TimeSeriesSplit":
            k_fold = TimeSeriesSplit(n_splits=self.n_splits)
            self.kfold = k_fold.split(self.train_data)

        elif kfold == "LeaveOneGroupOut":
            k_fold = LeaveOneGroupOut()
            groups = self._check_groups(groups=groups)
            self.kfold = k_fold.split(self.train_data, self.target_data, groups=groups)

        else:
            raise ValueError("cv need to be one of GroupKfold,StratifiedKFold,KFold,TimeSeriesSplit,LeaveOneGroupOut!")

    def _check_groups(self, groups):
        '''
        检验groups组数
        :param groups:
        :return:
        '''
        if groups.nunique() <= self.n_splits:
            self.n_splits = groups.nunique()
        return groups

    @staticmethod
    def get_ks_score(y_true, y_pred):
        d = zip(y_pred, y_true)
        bad = [k[0] for k in d if k[1] == 1]
        d = zip(y_pred, y_true)
        good = [k[0] for k in d if k[1] == 0]
        if len(bad) + len(good) < 100:
            return 0
        return ks_2samp(bad, good)[0]

    def cross_validation(self, params, kfold=None, groups=None, categorical_feature=None, n_splits=5):
        '''
        数据交叉验证
        :param params: dict, the params of estimator
        :param kfold: str, type of cv
        :param groups: ndarray or Series, the groups of split data
        :param categorical_feature: list or ndarray, the category feats of dataframe
        :param n_splits: int, the splits of data
        :return:
        '''
        if categorical_feature is None:
            categorical_feature = []

        if len(categorical_feature) > 0:
            self.train_data.loc[:, categorical_feature] = self.train_data.loc[:, categorical_feature].astype('category')

        self.n_splits = n_splits
        self._gen_kfold(kfold=kfold, groups=groups)

        oof = np.zeros(len(self.train_data))
        feature_importance_values = np.zeros(len(self.train_data.columns))
        valid_scores = []
        train_scores = []
        valid_ks_arr = []
        train_ks_arr = []

        for fold_, (trn_idx, val_idx) in enumerate(self.kfold):
            logger.info(f"start cv calulate, Fold: {fold_}")
            tr_x, tr_y = self.train_data.iloc[trn_idx, :], self.target_data[trn_idx]
            val_x, val_y = self.train_data.iloc[val_idx, :], self.target_data[val_idx]

            logger.info(f"train shape: {tr_x.shape}, train badRate: {tr_y.mean()}, "
                        f"valid shape: {val_x.shape}, valid badRate: {val_y.mean()}")

            model = lgb.LGBMClassifier().set_params(**params)

            if len(categorical_feature) == 0:
                model.fit(tr_x, tr_y, eval_metric='auc',
                          eval_set=[(val_x, val_y), (tr_x, tr_y)],
                          eval_names=['valid', 'train'],
                          early_stopping_rounds=100, verbose=100)
            else:
                model.fit(tr_x, tr_y, eval_metric='auc', categorical_feature=categorical_feature,
                          eval_set=[(val_x, val_y), (tr_x, tr_y)],eval_names=['valid', 'train'],
                          early_stopping_rounds=100, verbose=100)

            best_iteration = model.best_iteration_
            feature_importance_values += model.feature_importances_ / self.n_splits
            valid_prob = model.predict_proba(val_x, num_iteration=best_iteration)[:, 1]
            train_prob = model.predict_proba(tr_x, num_iteration=best_iteration)[:, 1]
            oof[val_idx] = valid_prob
            valid_score = model.best_score_['valid']['auc']
            train_score = model.best_score_['train']['auc']
            valid_ks = self.get_ks_score(val_y, valid_prob)
            train_ks = self.get_ks_score(tr_y, train_prob)

            valid_scores.append(valid_score)
            train_scores.append(train_score)
            valid_ks_arr.append(valid_ks)
            train_ks_arr.append(train_ks)

        del tr_x, val_x
        gc.collect()

        feats_importances = pd.DataFrame({'feature': self.train_data.columns, 'importance': feature_importance_values})
        feats_importances = feats_importances.sort_values('importance', ascending=False).reset_index()
        feats_importances['importance_normalized'] = feats_importances['importance'] / feats_importances['importance'].sum()
        feats_importances['cumsum'] = feats_importances['importance_normalized'].cumsum()
        valid_auc = roc_auc_score(self.target_data, oof)
        valid_scores.append(valid_auc)
        train_scores.append(np.mean(train_scores))
        valid_ks_arr.append(np.mean(valid_ks_arr))
        train_ks_arr.append(np.mean(train_ks_arr))

        fold_names = list(range(self.n_splits))
        fold_names.append('avg_score')
        result = pd.DataFrame({'fold': fold_names,
                               'train_auc': train_scores,
                               'valid_auc': valid_scores,
                               'train_ks': train_ks_arr,
                               'valid_ks': valid_ks_arr
                               })
        logger.info("cross validation complete!")
        return feats_importances, result


class IncreaseCVSelector(CheckDataType):
    def __init__(self, train_data, target_data):
        '''
        :param train_data: pd.Dataframe
        :param target_data: ndarray
        '''
        super(IncreaseCVSelector, self).__init__(train_data, target_data)
        self.feats_importances = None
        self.train_data = self.format_train
        self.target_data = self.format_target


    def _sort_feats_importanes(self, feats_importances):
        '''
        特征重要性排序
        :param feats_importances: pd.DataFrame
        :return:
        '''
        self.feats_importances = feats_importances.sort_values('importance', ascending=False)

    @staticmethod
    def _gen_cv_result(params, train_set, folds):
        '''
        cv模型初始化
        :param params: dict, the params of lgb
        :param train_set: DataSet, the Dataset of train
        :param folds: the type of split date
        :return: dict, the cv result
        '''
        cv_result = lgb.cv(params=params, train_set=train_set, folds=folds, shuffle=True)
        return cv_result

    def get_lgb_cv_score(self, feats_importances, total_iter, step, incre_params=None,
                         categorical_feature=None, auc_interval=None):
        '''
        模型通过cv自动筛选特征
        :param feats_importances: pd.Dataframe
        :param total_iter: int, the max iterator of traing round
        :param step: int, the numbers of add features
        :param incre_params: dict, the params of increase lgb model
        :param categorical_feature: list, the category of features
        :param auc_interval: int, the miniumn interval of auc increase
        :return:
            result: pd.DataFrame, the every step of cv result
            record_increase_feats: pd.DataFrame, the seleted of feats
        '''
        if categorical_feature is None:
            categorical_feature = []
        if auc_interval is None:
            auc_interval = 0.001
        if incre_params is None:
            incre_params = {
                "max_depth": 3,
                "learning_rate": 0.1,
                "num_boost_round": 600,
                "metrics": "auc",
                "verbose_eval": 200,
                "verbose":-1,
                "early_stopping_rounds": 100,
                "seed": 10
            }

        self._sort_feats_importanes(feats_importances=feats_importances)
        folds = StratifiedKFold(n_splits=5, shuffle=True, random_state=10)
        val_auc_arr = []
        val_auc_diff = []
        increase_feats = []
        feats_add_arr = []
        auc_initial = 0.5
        for i in range(0, min(len(self.feats_importances), total_iter), step):
            feats_add = self.feats_importances['feature'][i:i+step].values.tolist()
            logger.info(f"step:{i}, feature cnt:{len(increase_feats+feats_add)}")
            category_cols = [i for i in increase_feats+feats_add if i in categorical_feature]

            if len(category_cols) == 0:
                train_set = lgb.Dataset(self.train_data.loc[:, feats_add+increase_feats], label=self.target_data)
            else:
                train_set = lgb.Dataset(self.train_data.loc[:, feats_add+increase_feats], label=self.target_data,
                                        categorical_feature=category_cols)
            cv_result = self._gen_cv_result(params=incre_params, train_set=train_set, folds=folds)
            val_score = np.max(cv_result['auc-mean'])
            val_auc_arr.append(val_score)
            val_auc_diff.append(val_score - auc_initial)
            feats_add_arr.append(feats_add)

            if val_score - auc_initial < auc_interval:
                logger.info(f"delete the step features: {feats_add}, auc diff: {val_score - auc_initial}")
                continue

            increase_feats.extend(feats_add)
            auc_initial = val_score
            logger.info(f"add the step features, val_auc: {val_score}, val_std: {np.max(cv_result['auc-stdv'])}")
        result = pd.DataFrame(data={'step': range(step, min(len(self.feats_importances), total_iter)+step, step),
                                    'val_score': val_auc_arr, 'val_score_diff': val_auc_diff,
                                    'feats_add': feats_add_arr})
        record_increase_feats = self.feats_importances[self.feats_importances.feature.isin(increase_feats)]\
                                    .reset_index(drop=True)
        return result, record_increase_feats

